"""
Surface module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from __future__ import annotations
import uuid
import numpy as np
from typing import List, Union
from hub.city_model_structure.attributes.polygon import Polygon
from hub.city_model_structure.attributes.plane import Plane
from hub.city_model_structure.attributes.point import Point
from hub.city_model_structure.greenery.vegetation import Vegetation
from hub.city_model_structure.building_demand.thermal_boundary import ThermalBoundary
import hub.helpers.constants as cte


class Surface:
  """
  Surface class
  """

  def __init__(self, solid_polygon, perimeter_polygon, holes_polygons=None, name=None, surface_type=None):
    self._type = surface_type
    self._name = name
    self._id = None
    self._azimuth = None
    self._inclination = None
    self._area = None
    self._lower_corner = None
    self._upper_corner = None
    self._shared_surfaces = []
    self._global_irradiance = dict()
    self._perimeter_polygon = perimeter_polygon
    self._holes_polygons = holes_polygons
    self._solid_polygon = solid_polygon
    self._short_wave_reflectance = None
    self._long_wave_emittance = None
    self._inverse = None
    self._associated_thermal_boundaries = []
    self._vegetation = None

  @property
  def name(self):
    """
    Get the surface name
    :return: str
    """
    if self._name is None:
      self._name = str(uuid.uuid4())
    return self._name

  @property
  def id(self):
    """
    Get the surface id
    :return: str
    """
    if self._id is None:
      raise ValueError('Undefined surface id')
    return self._id

  @id.setter
  def id(self, value):
    """
    Set the surface id
    :param value: str
    """
    if value is not None:
      self._id = str(value)

  # todo: implement share surfaces
  @property
  def share_surfaces(self):
    """
    Raises not implemented error
    """
    raise NotImplementedError

  def _max_coord(self, axis):
    if axis == 'x':
      axis = 0
    elif axis == 'y':
      axis = 1
    else:
      axis = 2
    max_coordinate = ''
    for point in self.perimeter_polygon.coordinates:
      if max_coordinate == '':
        max_coordinate = point[axis]
      elif max_coordinate < point[axis]:
        max_coordinate = point[axis]
    return max_coordinate

  def _min_coord(self, axis):
    if axis == 'x':
      axis = 0
    elif axis == 'y':
      axis = 1
    else:
      axis = 2
    min_coordinate = ''
    for point in self.perimeter_polygon.coordinates:
      if min_coordinate == '':
        min_coordinate = point[axis]
      elif min_coordinate > point[axis]:
        min_coordinate = point[axis]
    return min_coordinate

  @property
  def lower_corner(self):
    """
    Get surface's lower corner [x, y, z]
    :return: [float]
    """
    if self._lower_corner is None:
      self._lower_corner = [self._min_coord('x'), self._min_coord('y'), self._min_coord('z')]
    return self._lower_corner

  @property
  def upper_corner(self):
    """
    Get surface's upper corner [x, y, z]
    :return: [float]
    """
    if self._upper_corner is None:
      self._upper_corner = [self._max_coord('x'), self._max_coord('y'), self._max_coord('z')]
    return self._upper_corner

  @property
  def perimeter_area(self):
    """
    Get perimeter surface area in square meters (opaque + transparent)
    :return: float
    """
    self._area = self.perimeter_polygon.area
    return self._area

  @property
  def azimuth(self):
    """
    Get surface azimuth in radians
    :return: float
    """
    if self._azimuth is None:
      normal = self.perimeter_polygon.normal
      self._azimuth = np.arctan2(normal[1], normal[0])
    return self._azimuth

  @property
  def inclination(self):
    """
    Get surface inclination in radians
    :return: float
    """
    if self._inclination is None:
      self._inclination = np.arccos(self.perimeter_polygon.normal[2])
    return self._inclination

  @property
  def type(self):
    """
    Get surface type Ground, Wall or Roof
    :return: str
    """

    # todo: there are more types: internal wall, internal floor... this method must be redefined
    if self._type is None:
      grad = np.rad2deg(self.inclination)
      if grad >= 170:
        self._type = 'Ground'
      elif 80 <= grad <= 100:
        self._type = 'Wall'
      else:
        self._type = 'Roof'
    return self._type

  @property
  def global_irradiance(self) -> dict:
    """
    Get global irradiance on surface in Wh/m2
    :return: dict{DataFrame(float)}
    """
    return self._global_irradiance

  @global_irradiance.setter
  def global_irradiance(self, value):
    """
    Set global irradiance on surface in Wh/m2
    :param value: dict{DataFrame(float)}
    """
    self._global_irradiance = value

  @property
  def perimeter_polygon(self) -> Polygon:
    """
    Get a polygon surface defined by the perimeter, merging solid and holes
    :return: Polygon
    """
    return self._perimeter_polygon

  @property
  def solid_polygon(self) -> Polygon:
    """
    Get the solid surface
    :return: Polygon
    """
    return self._solid_polygon

  @solid_polygon.setter
  def solid_polygon(self, value):
    """
    Set the solid surface
    :return: Polygon
    """
    self._solid_polygon = value

  @property
  def holes_polygons(self) -> Union[List[Polygon], None]:
    """
    Get hole surfaces, a list of hole polygons found in the surface
    :return: None, [] or [Polygon]
    None -> not known whether holes exist in reality or not due to low level of detail of input data
    [] -> no holes in the surface
    [Polygon] -> one or more holes in the surface
    """
    return self._holes_polygons

  @holes_polygons.setter
  def holes_polygons(self, value):
    """
    Set the hole surfaces
    :param value: [Polygon]
    """
    self._holes_polygons = value

  @property
  def short_wave_reflectance(self):
    """
    Get the short wave reflectance, this includes all solar spectrum, visible and not visible
    The absorptance as an opaque surface, can be calculated as 1-short_wave_reflectance
    :return: float
    """
    return self._short_wave_reflectance

  @short_wave_reflectance.setter
  def short_wave_reflectance(self, value):
    """
    Set the short wave reflectance, this includes all solar spectrum, visible and not visible
    The absorptance as an opaque surface, can be calculated as 1-short_wave_reflectance
    :param value: float
    """
    self._short_wave_reflectance = value

  @property
  def long_wave_emittance(self):
    """
    Get the long wave emittance af the surface
    The thermal absorptance can be calculated as 1-long_wave_emittance
    :return: float
    """
    return self._long_wave_emittance

  @long_wave_emittance.setter
  def long_wave_emittance(self, value):
    """
    Set the long wave emittance af the surface
    The thermal absorptance can be calculated as 1-long_wave_emittance
    :param value: float
    """
    self._long_wave_emittance = value

  @property
  def inverse(self) -> Surface:
    """
    Get the inverse surface (the same surface pointing backwards)
    :return: Surface
    """
    if self._inverse is None:
      new_solid_polygon = Polygon(self.solid_polygon.inverse)
      new_perimeter_polygon = Polygon(self.perimeter_polygon.inverse)
      new_holes_polygons = []
      if self.holes_polygons is not None:
        for hole in self.holes_polygons:
          new_holes_polygons.append(Polygon(hole.inverse))
      else:
        new_holes_polygons = None
      self._inverse = Surface(new_solid_polygon, new_perimeter_polygon, new_holes_polygons, cte.VIRTUAL_INTERNAL)
    return self._inverse

  def shared_surfaces(self):
    """
    Raises not implemented error
    """
    # todo: check https://trimsh.org/trimesh.collision.html as an option to implement this method
    raise NotImplementedError

  def divide(self, z):
    """
    Divides a surface at Z plane
    :return: Surface, Surface, Any
    """
    # todo: check return types
    # todo: recheck this method for LoD3 (windows)
    origin = Point([0, 0, z])
    normal = np.array([0, 0, 1])
    plane = Plane(normal=normal, origin=origin)
    polygon = self.perimeter_polygon
    part_1, part_2, intersection = polygon.divide(plane)
    surface_child = Surface(part_1, part_1, name=self.name, surface_type=self.type)
    rest_surface = Surface(part_2, part_2, name=self.name, surface_type=self.type)
    return surface_child, rest_surface, intersection

  @property
  def vegetation(self) -> Union[None, Vegetation]:
    """
    Get the vegetation construction at the external surface of the thermal boundary
    :return: None or Vegetation
    """
    return self._vegetation

  @vegetation.setter
  def vegetation(self, value):
    """
    Set the vegetation construction at the external surface of the thermal boundary
    :param value: Vegetation
    """
    self._vegetation = value

  @property
  def associated_thermal_boundaries(self) -> Union[None, List[ThermalBoundary]]:
    """
    Get the list of thermal boundaries that has this surface as external face
    :return: None or [ThermalBoundary]
    """
    return self._associated_thermal_boundaries

  @associated_thermal_boundaries.setter
  def associated_thermal_boundaries(self, value):
    """
    Set the list of thermal boundaries that has this surface as external face
    :param value: None or [ThermalBoundary]
    """
    self._associated_thermal_boundaries = value



"""
Polyhedron module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import List, Union
import sys
import math
import numpy as np
from trimesh import Trimesh
from hub.helpers.configuration_helper import ConfigurationHelper


class Polyhedron:
  """
  Polyhedron class
  """

  def __init__(self, polygons):
    self._polygons = polygons
    self._polyhedron = None
    self._triangulated_polyhedron = None
    self._volume = None
    self._faces = None
    self._vertices = None
    self._trimesh = None
    self._centroid = None
    self._max_z = None
    self._max_y = None
    self._max_x = None
    self._min_z = None
    self._min_y = None
    self._min_x = None

  def _position_of(self, point, face):
    """
    position of a specific point in the list of points that define a face
    :return: int
    """
    vertices = self.vertices
    for i in range(len(vertices)):
      # ensure not duplicated vertex
      power = 0
      vertex2 = vertices[i]
      for dimension in range(0, 3):
        power += math.pow(vertex2[dimension] - point[dimension], 2)
      distance = math.sqrt(power)
      if i not in face and distance == 0:
        return i
    return -1

  @property
  def vertices(self) -> np.ndarray:
    """
    Get polyhedron vertices
    :return: np.ndarray(int)
    """
    if self._vertices is None:
      vertices, self._vertices = [], []
      _ = [vertices.extend(s.coordinates) for s in self._polygons]
      for vertex_1 in vertices:
        found = False
        for vertex_2 in self._vertices:
          found = False
          power = 0
          for dimension in range(0, 3):
            power += math.pow(vertex_2[dimension] - vertex_1[dimension], 2)
          distance = math.sqrt(power)
          if distance == 0:
            found = True
            break
        if not found:
          self._vertices.append(vertex_1)
      self._vertices = np.asarray(self._vertices)
    return self._vertices

  @property
  def faces(self) -> List[List[int]]:
    """
    Get polyhedron triangular faces
    :return: [face]
    """
    if self._faces is None:
      self._faces = []

      for polygon in self._polygons:

        face = []
        points = polygon.coordinates
        if len(points) != 3:
          sub_polygons = polygon.triangulate()
          # todo: I modified this! To be checked @Guille
          if len(sub_polygons) >= 1:
            for sub_polygon in sub_polygons:
              face = []
              points = sub_polygon.coordinates
              for point in points:
                face.append(self._position_of(point, face))
              self._faces.append(face)
        else:
          for point in points:
            face.append(self._position_of(point, face))
          self._faces.append(face)
    return self._faces

  @property
  def trimesh(self) -> Union[Trimesh, None]:
    """
    Get polyhedron trimesh
    :return: Trimesh
    """
    if self._trimesh is None:
      for face in self.faces:
        if len(face) != 3:
          sys.stderr.write('Not able to generate trimesh\n')
          return None
      self._trimesh = Trimesh(vertices=self.vertices, faces=self.faces)
    return self._trimesh

  @property
  def volume(self):
    """
    Get polyhedron volume in cubic meters
    :return: float
    """
    if self._volume is None:
      if self.trimesh is None:
        self._volume = np.inf
      elif not self.trimesh.is_volume:
        self._volume = np.inf
      else:
        self._volume = self.trimesh.volume
    return self._volume

  @property
  def max_z(self):
    """
    Get polyhedron maximal z value in meters
    :return: float
    """
    if self._max_z is None:
      self._max_z = ConfigurationHelper().min_coordinate
      for polygon in self._polygons:
        for point in polygon.coordinates:
          self._max_z = max(self._max_z, point[2])
    return self._max_z

  @property
  def max_y(self):
    """
    Get polyhedron maximal y value in meters
    :return: float
    """
    if self._max_y is None:
      self._max_y = ConfigurationHelper().min_coordinate
      for polygon in self._polygons:
        for point in polygon.coordinates:
          if self._max_y < point[1]:
            self._max_y = point[1]
    return self._max_y

  @property
  def max_x(self):
    """
    Get polyhedron maximal x value in meters
    :return: float
    """
    if self._max_x is None:
      self._max_x = ConfigurationHelper().min_coordinate
      for polygon in self._polygons:
        for point in polygon.coordinates:
          self._max_x = max(self._max_x, point[0])
    return self._max_x

  @property
  def min_z(self):
    """
    Get polyhedron minimal z value in meters
    :return: float
    """
    if self._min_z is None:
      self._min_z = self.max_z
      for polygon in self._polygons:
        for point in polygon.coordinates:
          if self._min_z > point[2]:
            self._min_z = point[2]
    return self._min_z

  @property
  def min_y(self):
    """
    Get polyhedron minimal y value in meters
    :return: float
    """
    if self._min_y is None:
      self._min_y = self.max_y
      for polygon in self._polygons:
        for point in polygon.coordinates:
          if self._min_y > point[1]:
            self._min_y = point[1]
    return self._min_y

  @property
  def min_x(self):
    """
    Get polyhedron minimal x value in meters
    :return: float
    """
    if self._min_x is None:
      self._min_x = self.max_x
      for polygon in self._polygons:
        for point in polygon.coordinates:
          if self._min_x > point[0]:
            self._min_x = point[0]
    return self._min_x

  @property
  def centroid(self) -> Union[None, List[float]]:
    """
    Get polyhedron centroid
    :return: [x,y,z]
    """
    if self._centroid is None:
      trimesh = self.trimesh
      if trimesh is None:
        return None
      self._centroid = self.trimesh.centroid
    return self._centroid

  def stl_export(self, full_path):
    """
    Export the polyhedron to stl given file
    :param full_path: str
    :return: None
    """
    self.trimesh.export(full_path, 'stl_ascii')

  def obj_export(self, full_path):
    """
    Export the polyhedron to obj given file
    :param full_path: str
    :return: None
    """
    self.trimesh.export(full_path, 'obj')

  def show(self):
    """
    Auxiliary function to render the polyhedron
    :return: None
    """
    self.trimesh.show()



"""
Vegetation class
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import List
from hub.city_model_structure.greenery.soil import Soil
from hub.city_model_structure.greenery.plant import Plant


class Vegetation:
  def __init__(self, name, soil, soil_thickness, plants):
    self._name = name
    self._management = None
    self._air_gap = None
    self._soil = soil
    self._soil_thickness = soil_thickness
    self._plants = plants

  @property
  def name(self):
    """
    Get vegetation name
    :return: string
    """
    return self._name

  @property
  def management(self):
    """
    Get management
    :return: string
    """
    return self._management

  @management.setter
  def management(self, value):
    """
    Set management
    :param value: string
    """
    self._management = value

  @property
  def air_gap(self):
    """
    Get air gap in m
    :return: float
    """
    return self._air_gap

  @air_gap.setter
  def air_gap(self, value):
    """
    Set air gap in m
    :param value: float
    """
    self._air_gap = value

  @property
  def soil(self) -> Soil:
    """
    Get soil
    :return: Soil
    """
    return self._soil

  @property
  def soil_thickness(self):
    """
    Get soil thickness in m
    :return: float
    """
    return self._soil_thickness

  @property
  def plants(self) -> List[Plant]:
    """
    Get list plants in the vegetation
    :return: List[Plant]
    """
    return self._plants



"""
ComnetUsageParameters extracts the usage properties from Comnet catalog and assigns to each building
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
import copy
import sys
import numpy

import hub.helpers.constants as cte
from hub.helpers.dictionaries import Dictionaries
from hub.city_model_structure.building_demand.usage import Usage
from hub.city_model_structure.building_demand.lighting import Lighting
from hub.city_model_structure.building_demand.occupancy import Occupancy
from hub.city_model_structure.building_demand.appliances import Appliances
from hub.city_model_structure.building_demand.thermal_control import ThermalControl
from hub.city_model_structure.attributes.schedule import Schedule
from hub.city_model_structure.building_demand.internal_gain import InternalGain
from hub.catalog_factories.usage_catalog_factory import UsageCatalogFactory


class ComnetUsageParameters:
  """
  ComnetUsageParameters class
  """
  def __init__(self, city, base_path):
    self._city = city
    self._path = base_path

  def enrich_buildings(self):
    """
    Returns the city with the usage parameters assigned to the buildings
    :return:
    """
    city = self._city
    comnet_catalog = UsageCatalogFactory('comnet').catalog
    for building in city.buildings:
      usage_name = Dictionaries().hub_usage_to_comnet_usage[building.function]
      try:
        archetype_usage = self._search_archetypes(comnet_catalog, usage_name)
      except KeyError:
        sys.stderr.write(f'Building {building.name} has unknown usage archetype for building function:'
                         f' {building.function}')
        return

      for internal_zone in building.internal_zones:
        if internal_zone.area is None:
          raise Exception('Internal zone area not defined, ACH cannot be calculated')
        if internal_zone.volume is None:
          raise Exception('Internal zone volume not defined, ACH cannot be calculated')
        if internal_zone.area <= 0:
          raise Exception('Internal zone area is zero, ACH cannot be calculated')
        volume_per_area = internal_zone.volume / internal_zone.area
        usage = Usage()
        usage.name = usage_name
        self._assign_values(usage, archetype_usage, volume_per_area)
        usage.percentage = 1
        self._calculate_reduced_values_from_extended_library(usage, archetype_usage)

        internal_zone.usages = [usage]

  @staticmethod
  def _search_archetypes(comnet_catalog, usage_name):
    comnet_archetypes = comnet_catalog.entries('archetypes').usages
    for building_archetype in comnet_archetypes:
      if str(usage_name) == str(building_archetype.name):
        return building_archetype
    raise KeyError('archetype not found')

  @staticmethod
  def _assign_values(usage, archetype, volume_per_area):
    # Due to the fact that python is not a typed language, the wrong object type is assigned to
    # usage.occupancy when writing usage.occupancy = archetype.occupancy.
    # Same happens for lighting and appliances. Therefore, this walk around has been done.
    usage.mechanical_air_change = archetype.ventilation_rate / volume_per_area \
                                  * cte.HOUR_TO_MINUTES * cte.MINUTES_TO_SECONDS
    _occupancy = Occupancy()
    _occupancy.occupancy_density = archetype.occupancy.occupancy_density
    _occupancy.sensible_radiative_internal_gain = archetype.occupancy.sensible_radiative_internal_gain
    _occupancy.latent_internal_gain = archetype.occupancy.latent_internal_gain
    _occupancy.sensible_convective_internal_gain = archetype.occupancy.sensible_convective_internal_gain
    _occupancy.occupancy_schedules = archetype.occupancy.schedules
    usage.occupancy = _occupancy
    _lighting = Lighting()
    _lighting.density = archetype.lighting.density
    _lighting.convective_fraction = archetype.lighting.convective_fraction
    _lighting.radiative_fraction = archetype.lighting.radiative_fraction
    _lighting.latent_fraction = archetype.lighting.latent_fraction
    _lighting.schedules = archetype.lighting.schedules
    usage.lighting = _lighting
    _appliances = Appliances()
    _appliances.density = archetype.appliances.density
    _appliances.convective_fraction = archetype.appliances.convective_fraction
    _appliances.radiative_fraction = archetype.appliances.radiative_fraction
    _appliances.latent_fraction = archetype.appliances.latent_fraction
    _appliances.schedules = archetype.appliances.schedules
    usage.appliances = _appliances
    _control = ThermalControl()
    _control.cooling_set_point_schedules = archetype.thermal_control.cooling_set_point_schedules
    _control.heating_set_point_schedules = archetype.thermal_control.heating_set_point_schedules
    _control.hvac_availability_schedules = archetype.thermal_control.hvac_availability_schedules
    usage.thermal_control = _control

  @staticmethod
  def _calculate_reduced_values_from_extended_library(usage, archetype):
    number_of_days_per_type = {'WD': 251, 'Sat': 52, 'Sun': 62}
    total = 0
    for schedule in archetype.thermal_control.hvac_availability_schedules:
      if schedule.day_types[0] == cte.SATURDAY:
        for value in schedule.values:
          total += value * number_of_days_per_type['Sat']
      elif schedule.day_types[0] == cte.SUNDAY:
        for value in schedule.values:
          total += value * number_of_days_per_type['Sun']
      else:
        for value in schedule.values:
          total += value * number_of_days_per_type['WD']

    usage.hours_day = total / 365
    usage.days_year = 365

    max_heating_setpoint = cte.MIN_FLOAT
    min_heating_setpoint = cte.MAX_FLOAT

    for schedule in archetype.thermal_control.heating_set_point_schedules:
      if schedule.values is None:
        max_heating_setpoint = None
        min_heating_setpoint = None
        break
      if max(schedule.values) > max_heating_setpoint:
        max_heating_setpoint = max(schedule.values)
      if min(schedule.values) < min_heating_setpoint:
        min_heating_setpoint = min(schedule.values)

    min_cooling_setpoint = cte.MAX_FLOAT
    for schedule in archetype.thermal_control.cooling_set_point_schedules:
      if schedule.values is None:
        min_cooling_setpoint = None
        break
      if min(schedule.values) < min_cooling_setpoint:
        min_cooling_setpoint = min(schedule.values)

    usage.thermal_control.mean_heating_set_point = max_heating_setpoint
    usage.thermal_control.heating_set_back = min_heating_setpoint
    usage.thermal_control.mean_cooling_set_point = min_cooling_setpoint

  @staticmethod
  def _calculate_internal_gains(archetype):

    _DAYS = [cte.MONDAY, cte.TUESDAY, cte.WEDNESDAY, cte.THURSDAY, cte.FRIDAY, cte.SATURDAY, cte.SUNDAY, cte.HOLIDAY]
    _number_of_days_per_type = [51, 50, 50, 50, 50, 52, 52, 10]

    _mean_internal_gain = InternalGain()
    _mean_internal_gain.type = 'mean_value_of_internal_gains'
    _base_schedule = Schedule()
    _base_schedule.type = cte.INTERNAL_GAINS
    _base_schedule.time_range = cte.DAY
    _base_schedule.time_step = cte.HOUR
    _base_schedule.data_type = cte.FRACTION

    _latent_heat_gain = archetype.occupancy.latent_internal_gain
    _convective_heat_gain = archetype.occupancy.sensible_convective_internal_gain
    _radiative_heat_gain = archetype.occupancy.sensible_radiative_internal_gain
    _total_heat_gain = (_latent_heat_gain + _convective_heat_gain + _radiative_heat_gain)

    _schedule_values = numpy.zeros([24, 8])
    _sum = 0
    for day, _schedule in enumerate(archetype.occupancy.schedules):
      for v, value in enumerate(_schedule.values):
        _schedule_values[v, day] = value * _total_heat_gain
        _sum += value * _total_heat_gain * _number_of_days_per_type[day]

    _total_heat_gain += archetype.lighting.density + archetype.appliances.density
    _latent_heat_gain += archetype.lighting.latent_fraction * archetype.lighting.density\
                         + archetype.appliances.latent_fraction * archetype.appliances.density
    _radiative_heat_gain = archetype.lighting.radiative_fraction * archetype.lighting.density \
                           + archetype.appliances.radiative_fraction * archetype.appliances.density
    _convective_heat_gain = archetype.lighting.convective_fraction * archetype.lighting.density \
                           + archetype.appliances.convective_fraction * archetype.appliances.density

    for day, _schedule in enumerate(archetype.lighting.schedules):
      for v, value in enumerate(_schedule.values):
        _schedule_values[v, day] += value * archetype.lighting.density
        _sum += value * archetype.lighting.density * _number_of_days_per_type[day]

    for day, _schedule in enumerate(archetype.appliances.schedules):
      for v, value in enumerate(_schedule.values):
        _schedule_values[v, day] += value * archetype.appliances.density
        _sum += value * archetype.appliances.density * _number_of_days_per_type[day]

    _latent_fraction = _latent_heat_gain / _total_heat_gain
    _radiative_fraction = _radiative_heat_gain / _total_heat_gain
    _convective_fraction = _convective_heat_gain / _total_heat_gain
    _average_internal_gain = _sum / _total_heat_gain

    _schedules = []
    for day in range(0, len(_DAYS)):
      _schedule = copy.deepcopy(_base_schedule)
      _schedule.day_types = [_DAYS[day]]
      _schedule.values = _schedule_values[:day]
      _schedules.append(_schedule)

    _mean_internal_gain.average_internal_gain = _average_internal_gain
    _mean_internal_gain.latent_fraction = _latent_fraction
    _mean_internal_gain.convective_fraction = _convective_fraction
    _mean_internal_gain.radiative_fraction = _radiative_fraction
    _mean_internal_gain.schedules = _schedules

    return [_mean_internal_gain]



"""
UsageFactory retrieve the specific usage module for the given region
This factory can only be called after calling the construction factory so the thermal zones are created.
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from pathlib import Path
from hub.imports.usage.comnet_usage_parameters import ComnetUsageParameters
from hub.imports.usage.nrcan_usage_parameters import NrcanUsageParameters
from hub.hub_logger import logger
from hub.helpers.utils import validate_import_export_type


class UsageFactory:
  """
  UsageFactory class
  """
  def __init__(self, handler, city, base_path=None):
    if base_path is None:
      base_path = Path(Path(__file__).parent.parent / 'data/usage')
    self._handler = '_' + handler.lower().replace(' ', '_')
    class_funcs = validate_import_export_type(UsageFactory)
    if self._handler not in class_funcs:
      err_msg = f"Wrong import type [{self._handler}]. Valid functions include {class_funcs}"
      logger.error(err_msg)
      raise Exception(err_msg)
    self._city = city
    self._base_path = base_path

  def _comnet(self):
    """
    Enrich the city with COMNET usage library
    """
    self._city.level_of_detail.usage = 2
    ComnetUsageParameters(self._city, self._base_path).enrich_buildings()

  def _nrcan(self):
    """
    Enrich the city with NRCAN usage library
    """
    self._city.level_of_detail.usage = 2
    NrcanUsageParameters(self._city, self._base_path).enrich_buildings()

  def enrich(self):
    """
    Enrich the city given to the class using the usage factory given handler
    :return: None
    """
    getattr(self, self._handler, lambda: None)()



"""
Connection module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Guille guille.gutierrezmorote@concordia.ca
"""

import ast
from typing import Union

from hub.city_model_structure.attributes.edge import Edge
from hub.city_model_structure.transport.lane import Lane


class Connection:
  """
  Connection class
  """

  def __init__(self):
    self._from_edge = None
    self._to_edge = None
    self._from_lane = None
    self._to_lane = None
    self._pass = None
    self._keep_clear = None

  @property
  def from_edge(self) -> Edge:
    """
    Get "from" edge
    :return: Edge
    """
    return self._from_edge

  @from_edge.setter
  def from_edge(self, value):
    """
    Set "from" edge
    :param value: Edge
    """
    self._from_edge = value

  @property
  def to_edge(self) -> Edge:
    """
    Get "to" edge
    :return: Edge
    """
    return self._to_edge

  @to_edge.setter
  def to_edge(self, value):
    """
    Set "to" edge
    :param value: Edge
    """
    self._to_edge = value

  @property
  def from_lane(self) -> Lane:
    """
    Get "from" lane
    :return: Lane
    """
    return self._to_lane

  @from_lane.setter
  def from_lane(self, value):
    """
    Set "from" lane
    :param value: Lane
    """
    self._from_lane = value

  @property
  def to_lane(self) -> Lane:
    """
    Get "to" lane
    :return: Lane
    """
    return self._to_lane

  @to_lane.setter
  def to_lane(self, value):
    """
    Set "to" lane
    :param value: Lane
    """
    self._to_lane = value

  @property
  def pass_not_wait(self) -> Union[None, bool]:
    """
    Get if the vehicles which pass this (lane to lane) connection will not wait
    :return: None or Boolean
    """
    return self._pass

  @pass_not_wait.setter
  def pass_not_wait(self, value):
    """
    Set if the vehicles which pass this (lane to lane) connection will not wait
    :param value: Boolean
    """
    if value is not None:
      self._pass = ast.literal_eval(value)

  @property
  def keep_clear(self) -> Union[None, bool]:
    """
    Get if vehicles which pass this (lane to lane) connection should keep the intersection clear
    :return: None or Boolean
    """
    return self._keep_clear

  @keep_clear.setter
  def keep_clear(self, value):
    """
    Set if vehicles which pass this (lane to lane) connection should keep the intersection clear
    :param value: Boolean
    """
    if value is not None:
      self._keep_clear = ast.literal_eval(value)



"""
Idf exports one building to idf format
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Guillermo.GutierrezMorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
                   Soroush Samareh Abolhassani soroush.samarehabolhassani@mail.concordia.ca
"""
import copy
from pathlib import Path
from geomeppy import IDF
import hub.helpers.constants as cte
from hub.city_model_structure.attributes.schedule import Schedule


class Idf:
  """
  Exports city to IDF
  """
  _BUILDING = 'BUILDING'
  _ZONE = 'ZONE'
  _LIGHTS = 'LIGHTS'
  _PEOPLE = 'PEOPLE'
  _THERMOSTAT = 'HVACTEMPLATE:THERMOSTAT'
  _IDEAL_LOAD_AIR_SYSTEM = 'HVACTEMPLATE:ZONE:IDEALLOADSAIRSYSTEM'
  _SURFACE = 'BUILDINGSURFACE:DETAILED'
  _SHADING = 'SHADING:BUILDING:DETAILED'
  _SHADING_PROPERTY = 'SHADINGPROPERTY:REFLECTANCE'
  _BUILDING_SURFACE = 'BuildingSurfaceDetailed'
  _CONSTRUCTION = 'CONSTRUCTION'
  _MATERIAL = 'MATERIAL'
  _MATERIAL_NOMASS = 'MATERIAL:NOMASS'
  _MATERIAL_ROOFVEGETATION = 'MATERIAL:ROOFVEGETATION'

  _WINDOW = 'FENESTRATIONSURFACE:DETAILED'
  _WINDOW_MATERIAL_SIMPLE = 'WINDOWMATERIAL:SIMPLEGLAZINGSYSTEM'
  _ROUGHNESS = 'MediumRough'
  _INFILTRATION = 'ZONEINFILTRATION:DESIGNFLOWRATE'

  _HOURLY_SCHEDULE = 'SCHEDULE:DAY:HOURLY'
  _COMPACT_SCHEDULE = 'SCHEDULE:COMPACT'
  _FILE_SCHEDULE = 'SCHEDULE:FILE'
  _SCHEDULE_LIMIT = 'SCHEDULETYPELIMITS'
  _ON_OFF = 'On/Off'
  _FRACTION = 'Fraction'
  _ANY_NUMBER = 'Any Number'
  _CONTINUOUS = 'Continuous'
  _DISCRETE = 'Discrete'
  _SIZING_PERIODS = 'SIZINGPERIOD:DESIGNDAY'
  _LOCATION = 'SITE:LOCATION'
  _SIMPLE = 'Simple'

  idf_surfaces = {
    # todo: make an enum for all the surface types
    cte.WALL: 'wall',
    cte.GROUND: 'floor',
    cte.ROOF: 'roof'
  }
  idf_type_limits = {
    cte.ON_OFF: 'on/off',
    cte.FRACTION: 'Fraction',
    cte.ANY_NUMBER: 'Any Number',
    cte.CONTINUOUS: 'Continuous',
    cte.DISCRETE: 'Discrete'
  }
  idf_day_types = {
    cte.MONDAY: 'Monday',
    cte.TUESDAY: 'Tuesday',
    cte.WEDNESDAY: 'Wednesday',
    cte.THURSDAY: 'Thursday',
    cte.FRIDAY: 'Friday',
    cte.SATURDAY: 'Saturday',
    cte.SUNDAY: 'Sunday',
    cte.HOLIDAY: 'Holidays',
    cte.WINTER_DESIGN_DAY: 'WinterDesignDay',
    cte.SUMMER_DESIGN_DAY: 'SummerDesignDay'
  }

  def __init__(self, city, output_path, idf_file_path, idd_file_path, epw_file_path, export_type="Surfaces",
               target_buildings=None, adjacent_buildings=None):
    self._city = city
    self._output_path = str(output_path.resolve())
    self._output_file = str((output_path / f'{city.name}.idf').resolve())
    self._export_type = export_type
    self._idd_file_path = str(idd_file_path)
    self._idf_file_path = str(idf_file_path)
    self._epw_file_path = str(epw_file_path)
    IDF.setiddname(self._idd_file_path)
    self._idf = IDF(self._idf_file_path, self._epw_file_path)
    self._idf.newidfobject(self._SCHEDULE_LIMIT, Name=self._ANY_NUMBER)
    self._idf.newidfobject(self._SCHEDULE_LIMIT, Name=self._FRACTION, Lower_Limit_Value=0.0, Upper_Limit_Value=1.0,
                           Numeric_Type=self._CONTINUOUS)
    self._idf.newidfobject(self._SCHEDULE_LIMIT, Name=self._ON_OFF, Lower_Limit_Value=0, Upper_Limit_Value=1,
                           Numeric_Type=self._DISCRETE)
    self._target_buildings = target_buildings
    if target_buildings is None:
      self._target_buildings = [building.name for building in self._city.buildings]
    self._adjacent_buildings = adjacent_buildings
    if self._adjacent_buildings is None:
      self._adjacent_buildings = []
    self._export()

  @staticmethod
  def _matrix_to_list(points, lower_corner):
    lower_x = lower_corner[0]
    lower_y = lower_corner[1]
    lower_z = lower_corner[2]
    points_list = []
    for point in points:
      point_tuple = (point[0] - lower_x, point[1] - lower_y, point[2] - lower_z)
      points_list.append(point_tuple)
    return points_list

  @staticmethod
  def _matrix_to_2d_list(points):
    points_list = []
    for point in points:
      point_tuple = (point[0], point[1])
      points_list.append(point_tuple)
    return points_list

  def _add_material(self, layer):
    for material in self._idf.idfobjects[self._MATERIAL]:
      if material.Name == layer.material.name:
        return
    for material in self._idf.idfobjects[self._MATERIAL_NOMASS]:
      if material.Name == layer.material.name:
        return
    if layer.material.no_mass:
      self._idf.newidfobject(self._MATERIAL_NOMASS,
                             Name=layer.material.name,
                             Roughness=self._ROUGHNESS,
                             Thermal_Resistance=layer.material.thermal_resistance,
                             Thermal_Absorptance=layer.material.thermal_absorptance,
                             Solar_Absorptance=layer.material.solar_absorptance,
                             Visible_Absorptance=layer.material.visible_absorptance
                             )
    else:
      self._idf.newidfobject(self._MATERIAL,
                             Name=layer.material.name,
                             Roughness=self._ROUGHNESS,
                             Thickness=layer.thickness,
                             Conductivity=layer.material.conductivity,
                             Density=layer.material.density,
                             Specific_Heat=layer.material.specific_heat,
                             Thermal_Absorptance=layer.material.thermal_absorptance,
                             Solar_Absorptance=layer.material.solar_absorptance,
                             Visible_Absorptance=layer.material.visible_absorptance
                             )

  def _add_standard_compact_hourly_schedule(self, usage, schedule_type, schedules):
    for schedule in self._idf.idfobjects[self._COMPACT_SCHEDULE]:
      if schedule.Name == f'{schedule_type} schedules {usage}':
        return
    _kwargs = {'Name': f'{schedule_type} schedules {usage}',
               'Schedule_Type_Limits_Name': self.idf_type_limits[schedules[0].data_type],
               'Field_1': 'Through: 12/31'}
    counter = 1
    for j, schedule in enumerate(schedules):
      _val = schedule.values
      _new_field = ''
      for day_type in schedule.day_types:
        _new_field += f' {self.idf_day_types[day_type]}'
      _kwargs[f'Field_{j * 25 + 2}'] = f'For:{_new_field}'
      counter += 1
      for i in range(0, len(_val)):
        _kwargs[f'Field_{j * 25 + 3 + i}'] = f'Until: {i + 1:02d}:00,{_val[i]}'
        counter += 1
    _kwargs[f'Field_{counter + 1}'] = f'For AllOtherDays'
    _kwargs[f'Field_{counter + 2}'] = f'Until: 24:00,0.0'
    self._idf.newidfobject(self._COMPACT_SCHEDULE, **_kwargs)

  def _write_schedules_file(self, usage, schedule):
    file_name = str((Path(self._output_path) / f'{schedule.type} schedules {usage}.dat').resolve())
    with open(file_name, 'w') as file:
      for value in schedule.values:
        file.write(f'{str(value)},\n')
    return file_name

  def _add_file_schedule(self, usage, schedule, file_name):
    _schedule = self._idf.newidfobject(self._FILE_SCHEDULE, Name=f'{schedule.type} schedules {usage}')
    _schedule.Schedule_Type_Limits_Name = self.idf_type_limits[schedule.data_type]
    _schedule.File_Name = file_name
    _schedule.Column_Number = 1
    _schedule.Rows_to_Skip_at_Top = 0
    _schedule.Number_of_Hours_of_Data = 8760
    _schedule.Column_Separator = 'Comma'
    _schedule.Interpolate_to_Timestep = 'No'
    _schedule.Minutes_per_Item = 60

  def _add_infiltration_schedules(self, thermal_zone):
    _infiltration_schedules = []
    for hvac_availability_schedule in thermal_zone.thermal_control.hvac_availability_schedules:
      _schedule = Schedule()
      _schedule.type = cte.INFILTRATION
      _schedule.data_type = cte.FRACTION
      _schedule.time_step = cte.HOUR
      _schedule.time_range = cte.DAY
      _schedule.day_types = copy.deepcopy(hvac_availability_schedule.day_types)
      _infiltration_values = []
      for hvac_value in hvac_availability_schedule.values:
        if hvac_value == 0:
          _infiltration_values.append(thermal_zone.infiltration_rate_system_off)
        else:
          _infiltration_values.append(thermal_zone.infiltration_rate_system_on)
      _schedule.values = _infiltration_values
      _infiltration_schedules.append(_schedule)
    for schedule in self._idf.idfobjects[self._HOURLY_SCHEDULE]:
      if schedule.Name == f'Infiltration schedules {thermal_zone.usage_name}':
        return
    return self._add_standard_compact_hourly_schedule(thermal_zone.usage_name, 'Infiltration', _infiltration_schedules)

  def _add_people_activity_level_schedules(self, thermal_zone):
    _occ = thermal_zone.occupancy
    if _occ.occupancy_density == 0:
      _total_heat = 0
    else:
      _total_heat = (_occ.sensible_convective_internal_gain + _occ.sensible_radiative_internal_gain
                     + _occ.latent_internal_gain) / _occ.occupancy_density
    for schedule in self._idf.idfobjects[self._COMPACT_SCHEDULE]:
      if schedule.Name == f'Activity Level schedules {thermal_zone.usage_name}':
        return
    _kwargs = {'Name': f'Activity Level schedules {thermal_zone.usage_name}',
               'Schedule_Type_Limits_Name': self.idf_type_limits[cte.ANY_NUMBER],
               'Field_1': 'Through: 12/31',
               'Field_2': 'For AllDays',
               'Field_3': f'Until: 24:00,{_total_heat}'}
    self._idf.newidfobject(self._COMPACT_SCHEDULE, **_kwargs)
    return

  def _add_schedules(self, usage, schedule_type, new_schedules):
    schedule_from_file = False
    for schedule in new_schedules:
      if len(schedule.values) > 168:  # Hours in one week
        schedule_from_file = True
    if schedule_from_file:
      for schedule in self._idf.idfobjects[self._FILE_SCHEDULE]:
        if schedule.Name == f'{schedule_type} schedules {usage}':
          return
      file_name = self._write_schedules_file(usage, new_schedules[0])
      return self._add_file_schedule(usage, new_schedules[0], file_name)
    else:
      for schedule in self._idf.idfobjects[self._HOURLY_SCHEDULE]:
        if schedule.Name == f'{schedule_type} schedules {usage}':
          return
      return self._add_standard_compact_hourly_schedule(usage, schedule_type, new_schedules)

  def _add_construction(self, thermal_boundary):
    for construction in self._idf.idfobjects[self._CONSTRUCTION]:
      if thermal_boundary.parent_surface.vegetation is not None:
        if construction.Name == f'{thermal_boundary.construction_name}_{thermal_boundary.parent_surface.vegetation.name}':
          return
      else:
        if construction.Name == thermal_boundary.construction_name:
          return
    if thermal_boundary.layers is None:
      for material in self._idf.idfobjects[self._MATERIAL]:
        if material.Name == "DefaultMaterial":
          return
      self._idf.set_default_constructions()
      return
    for layer in thermal_boundary.layers:
      self._add_material(layer)
    layers = thermal_boundary.layers
    # The constructions should have at least one layer
    if thermal_boundary.parent_surface.vegetation is not None:
      _kwargs = {'Name': f'{thermal_boundary.construction_name}_{thermal_boundary.parent_surface.vegetation.name}',
                 'Outside_Layer': thermal_boundary.parent_surface.vegetation.name}
      for i in range(0, len(layers) - 1):
        _kwargs[f'Layer_{i + 2}'] = layers[i].material.name
    else:
      _kwargs = {'Name': thermal_boundary.construction_name, 'Outside_Layer': layers[0].material.name}
      for i in range(1, len(layers) - 1):
        _kwargs[f'Layer_{i + 1}'] = layers[i].material.name
    self._idf.newidfobject(self._CONSTRUCTION, **_kwargs)

  def _add_window_construction_and_material(self, thermal_opening):
    for window_material in self._idf.idfobjects[self._WINDOW_MATERIAL_SIMPLE]:
      if window_material['UFactor'] == thermal_opening.overall_u_value and \
         window_material['Solar_Heat_Gain_Coefficient'] == thermal_opening.g_value:
        return

    order = str(len(self._idf.idfobjects[self._WINDOW_MATERIAL_SIMPLE]) + 1)
    material_name = 'glazing_' + order
    _kwargs = {'Name': material_name, 'UFactor': thermal_opening.overall_u_value,
               'Solar_Heat_Gain_Coefficient': thermal_opening.g_value}
    self._idf.newidfobject(self._WINDOW_MATERIAL_SIMPLE, **_kwargs)

    window_construction_name = 'window_construction_' + order
    _kwargs = {'Name': window_construction_name, 'Outside_Layer': material_name}
    self._idf.newidfobject(self._CONSTRUCTION, **_kwargs)

  def _add_zone(self, thermal_zone, name):
    for zone in self._idf.idfobjects['ZONE']:
      if zone.Name == name:
        return
    # todo: what do we need to define a zone in energy plus?
    self._idf.newidfobject(self._ZONE, Name=name, Volume=thermal_zone.volume)
    self._add_heating_system(thermal_zone, name)

  def _add_thermostat(self, thermal_zone):
    thermostat_name = f'Thermostat {thermal_zone.usage_name}'
    for thermostat in self._idf.idfobjects[self._THERMOSTAT]:
      if thermostat.Name == thermostat_name:
        return thermostat
    # todo: change schedules to schedule name and create schedules using the add_schedule function
    return self._idf.newidfobject(self._THERMOSTAT,
                                  Name=thermostat_name,
                                  Heating_Setpoint_Schedule_Name=f'Heating thermostat schedules {thermal_zone.usage_name}',
                                  Cooling_Setpoint_Schedule_Name=f'Cooling thermostat schedules {thermal_zone.usage_name}')

  def _add_heating_system(self, thermal_zone, zone_name):
    for air_system in self._idf.idfobjects[self._IDEAL_LOAD_AIR_SYSTEM]:
      if air_system.Zone_Name == zone_name:
        return
    thermostat = self._add_thermostat(thermal_zone)
    self._idf.newidfobject(self._IDEAL_LOAD_AIR_SYSTEM,
                           Zone_Name=zone_name,
                           System_Availability_Schedule_Name=f'HVAC AVAIL SCHEDULES {thermal_zone.usage_name}',
                           Heating_Availability_Schedule_Name=f'HVAC AVAIL SCHEDULES {thermal_zone.usage_name}',
                           Cooling_Availability_Schedule_Name=f'HVAC AVAIL SCHEDULES {thermal_zone.usage_name}',
                           Template_Thermostat_Name=thermostat.Name)

  def _add_occupancy(self, thermal_zone, zone_name):
    number_of_people = thermal_zone.occupancy.occupancy_density * thermal_zone.total_floor_area
    fraction_radiant = thermal_zone.occupancy.sensible_radiative_internal_gain / \
                       (thermal_zone.occupancy.sensible_radiative_internal_gain +
                        thermal_zone.occupancy.sensible_convective_internal_gain)
    self._idf.newidfobject(self._PEOPLE,
                           Name=f'{zone_name}_occupancy',
                           Zone_or_ZoneList_Name=zone_name,
                           Number_of_People_Schedule_Name=f'Occupancy schedules {thermal_zone.usage_name}',
                           Number_of_People_Calculation_Method="People",
                           Number_of_People=number_of_people,
                           Fraction_Radiant=fraction_radiant,
                           Activity_Level_Schedule_Name=f'Activity Level schedules {thermal_zone.usage_name}'
                           )

  def _add_infiltration(self, thermal_zone, zone_name):
    for zone in self._idf.idfobjects["ZONE"]:
      if zone.Name == f'{zone_name}_infiltration':
        return
    self._idf.newidfobject(self._INFILTRATION,
                           Name=f'{zone_name}_infiltration',
                           Zone_or_ZoneList_Name=zone_name,
                           Schedule_Name=f'Infiltration schedules {thermal_zone.usage_name}',
                           Design_Flow_Rate_Calculation_Method='AirChanges/Hour',
                           Air_Changes_per_Hour=thermal_zone.mechanical_air_change
                           )

  def _rename_building(self, city_name):
    name = str(str(city_name.encode("utf-8")))
    for building in self._idf.idfobjects[self._BUILDING]:
      building.Name = f'Buildings in {name}'
      building['Solar_Distribution'] = 'FullExterior'

  def _remove_sizing_periods(self):
    while len(self._idf.idfobjects[self._SIZING_PERIODS]) > 0:
      self._idf.popidfobject(self._SIZING_PERIODS, 0)

  def _remove_location(self):
    self._idf.popidfobject(self._LOCATION, 0)

  def _export(self):
    """
    Export the idf file into the given path.

    If buildings to calculate are provided, only those will appear in the output variables, otherwise all the city
    buildings will be calculated.
    If adjacent buildings are provided those buildings will be calculated, but will not appear in the output variables.

    export type = "Surfaces|Block"
    """

    self._remove_location()
    self._remove_sizing_periods()
    self._rename_building(self._city.name)
    self._lod = self._city.level_of_detail.geometry
    for building in self._city.buildings:

      for internal_zone in building.internal_zones:
        for thermal_zone in internal_zone.thermal_zones:
          for thermal_boundary in thermal_zone.thermal_boundaries:
            self._add_construction(thermal_boundary)
            if thermal_boundary.parent_surface.vegetation is not None:
              self._add_vegetation_material(thermal_boundary.parent_surface.vegetation)
            for thermal_opening in thermal_boundary.thermal_openings:
              self._add_window_construction_and_material(thermal_opening)
          usage = thermal_zone.usage_name
          if building.name in self._target_buildings or building.name in self._adjacent_buildings:
            self._add_infiltration_schedules(thermal_zone)
            self._add_schedules(usage, 'Occupancy', thermal_zone.occupancy.occupancy_schedules)
            self._add_schedules(usage, 'HVAC AVAIL', thermal_zone.thermal_control.hvac_availability_schedules)
            self._add_schedules(usage, 'Heating thermostat', thermal_zone.thermal_control.heating_set_point_schedules)
            self._add_schedules(usage, 'Cooling thermostat', thermal_zone.thermal_control.cooling_set_point_schedules)
            self._add_people_activity_level_schedules(thermal_zone)

            self._add_zone(thermal_zone, building.name)
            self._add_heating_system(thermal_zone, building.name)
            self._add_infiltration(thermal_zone, building.name)
            self._add_occupancy(thermal_zone, building.name)
      if self._export_type == "Surfaces":
        if building.name in self._target_buildings or building.name in self._adjacent_buildings:
          self._add_surfaces(building, building.name)
        else:
          self._add_shading(building)
      else:
        self._add_block(building)
    # todo: this should change to specific variables per zone to process only the ones in the buildings_to_calculate
    for building in self._target_buildings:
      continue

    self._idf.newidfobject(
        "OUTPUT:VARIABLE",
        Variable_Name="Zone Ideal Loads Supply Air Total Heating Energy",
        Reporting_Frequency="Hourly",
        )

    self._idf.newidfobject(
        "OUTPUT:VARIABLE",
        Variable_Name="Zone Ideal Loads Supply Air Total Cooling Energy",
        Reporting_Frequency="Hourly",
        )
    self._idf.match()
    try:
      self._idf.intersect_match()
    except IndexError:
      # seems to be a bug from geomeppy when surfaces cannot be intersected
      pass

    # post-process to erase windows associated to adiabatic walls
    windows_list = []
    for window in self._idf.idfobjects[self._WINDOW]:
      found = False
      for surface in self._idf.idfobjects[self._SURFACE]:
        if window.Building_Surface_Name == surface.Name:
          found = True
      if not found:
        windows_list.append(window)
    for window in windows_list:
      self._idf.removeidfobject(window)

    self._idf.saveas(str(self._output_file))
    return self._idf

  def run(self):
    """
    Start the energy plus simulation
    """
    self._idf.run(expandobjects=False, readvars=True, output_directory=self._output_path,
                  output_prefix=f'{self._city.name}_')

  def _add_block(self, building):
    _points = self._matrix_to_2d_list(building.foot_print.coordinates)
    self._idf.add_block(name=building.name, coordinates=_points, height=building.max_height,
                        num_stories=int(building.storeys_above_ground))

    for surface in self._idf.idfobjects[self._SURFACE]:
      for thermal_zone in building.thermal_zones:
        for boundary in thermal_zone.thermal_boundaries:
          if surface.Type == self.idf_surfaces[boundary.surface.type]:
            surface.Construction_Name = boundary.construction_name
            break
        for usage in thermal_zone.usages:
          surface.Zone_Name = usage.id
          break
        break
    self._idf.intersect_match()

  def _add_shading(self, building):
    for surface in building.surfaces:
      shading = self._idf.newidfobject(self._SHADING, Name=f'{surface.name}')
      coordinates = self._matrix_to_list(surface.solid_polygon.coordinates,
                                         self._city.lower_corner)
      shading.setcoords(coordinates)
      solar_reflectance = surface.short_wave_reflectance
      self._idf.newidfobject(self._SHADING_PROPERTY,
                             Shading_Surface_Name=f'{surface.name}',
                             Diffuse_Solar_Reflectance_of_Unglazed_Part_of_Shading_Surface=solar_reflectance,
                             Fraction_of_Shading_Surface_That_Is_Glazed=0)

    # todo: Add properties for that name

  def _add_surfaces(self, building, zone_name):
    for internal_zone in building.internal_zones:
      for thermal_zone in internal_zone.thermal_zones:
        for boundary in thermal_zone.thermal_boundaries:
          idf_surface_type = self.idf_surfaces[boundary.parent_surface.type]
          outside_boundary_condition = 'Outdoors'
          sun_exposure = 'SunExposed'
          wind_exposure = 'WindExposed'
          if boundary.parent_surface.type == cte.GROUND:
            outside_boundary_condition = 'Ground'
            sun_exposure = 'NoSun'
            wind_exposure = 'NoWind'
          if boundary.parent_surface.vegetation is not None:
            construction_name = f'{boundary.construction_name}_{boundary.parent_surface.vegetation.name}'
          else:
            construction_name = boundary.construction_name
          surface = self._idf.newidfobject(self._SURFACE, Name=f'{boundary.parent_surface.name}',
                                           Surface_Type=idf_surface_type,
                                           Zone_Name=zone_name,
                                           Construction_Name=construction_name,
                                           Outside_Boundary_Condition=outside_boundary_condition,
                                           Sun_Exposure=sun_exposure,
                                           Wind_Exposure=wind_exposure)
          coordinates = self._matrix_to_list(boundary.parent_surface.solid_polygon.coordinates,
                                             self._city.lower_corner)

          surface.setcoords(coordinates)

    if self._lod >= 3:
      for internal_zone in building.internal_zones:
        for thermal_zone in internal_zone.thermal_zones:
          for boundary in thermal_zone.thermal_boundaries:
            self._add_windows_by_vertices(boundary)
    else:
      # idf only allows setting wwr for external walls
      wwr = 0
      for surface in building.surfaces:
        if surface.type == cte.WALL:
          wwr = surface.associated_thermal_boundaries[0].window_ratio
      self._idf.set_wwr(wwr, construction='window_construction_1')

  def _add_windows_by_vertices(self, boundary):
    raise NotImplementedError

  def _compare_window_constructions(self, window_construction, opening):
    glazing = window_construction['Outside_Layer']
    for material in self._idf.idfobjects[self._WINDOW_MATERIAL_SIMPLE]:
      if material['Name'] == glazing:
        if material['UFactor'] == opening.overall_u_value and \
           material['Solar_Heat_Gain_Coefficient'] == opening.g_value:
          return True
    return False

  def _add_vegetation_material(self, vegetation):
    for vegetation_material in self._idf.idfobjects[self._MATERIAL_ROOFVEGETATION]:
      if vegetation_material.Name == vegetation.name:
        return
    soil = vegetation.soil
    height = 0
    leaf_area_index = 0
    leaf_reflectivity = 0
    leaf_emissivity = 0
    minimal_stomatal_resistance = 0
    for plant in vegetation.plants:
      height += plant.percentage * plant.height
      leaf_area_index += plant.percentage * plant.leaf_area_index
      leaf_reflectivity += plant.percentage * plant.leaf_reflectivity
      leaf_emissivity += plant.percentage * plant.leaf_emissivity
      minimal_stomatal_resistance += plant.percentage * plant.minimal_stomatal_resistance
    self._idf.newidfobject(self._MATERIAL_ROOFVEGETATION,
                           Name=vegetation.name,
                           Height_of_Plants=height,
                           Leaf_Area_Index=leaf_area_index,
                           Leaf_Reflectivity=leaf_reflectivity,
                           Leaf_Emissivity=leaf_emissivity,
                           Minimum_Stomatal_Resistance=minimal_stomatal_resistance,
                           Soil_Layer_Name=soil.name,
                           Roughness=soil.roughness,
                           Thickness=vegetation.soil_thickness,
                           Conductivity_of_Dry_Soil=soil.dry_conductivity,
                           Density_of_Dry_Soil=soil.dry_density,
                           Specific_Heat_of_Dry_Soil=soil.dry_specific_heat,
                           Thermal_Absorptance=soil.thermal_absorptance,
                           Solar_Absorptance=soil.solar_absorptance,
                           Visible_Absorptance=soil.visible_absorptance,
                           Saturation_Volumetric_Moisture_Content_of_the_Soil_Layer=
                           soil.saturation_volumetric_moisture_content,
                           Residual_Volumetric_Moisture_Content_of_the_Soil_Layer=
                           soil.residual_volumetric_moisture_content,
                           Initial_Volumetric_Moisture_Content_of_the_Soil_Layer=
                           soil.initial_volumetric_moisture_content,
                           Moisture_Diffusion_Calculation_Method=self._SIMPLE)



"""
Occupant module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Sanam Dabirian sanam.dabirian@mail.concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""


class Occupant:
  """
  Occupant class
  """

  def __init__(self):
    """
    Constructor
    """

    self._heat_dissipation = None
    self._occupancy_rate = None
    self._occupant_type = None
    self._arrival_time = None
    self._departure_time = None
    self._break_time = None
    self._day_of_week = None
    self._pd_of_meetings_duration = None

  @property
  def heat_dissipation(self):
    """
    Get heat dissipation of occupants in W/person
    :return: float
    """
    return self._heat_dissipation

  @heat_dissipation.setter
  def heat_dissipation(self, value):
    """
    Set heat dissipation of occupants in W/person
    :param value: float
    """
    self._heat_dissipation = float(value)

  @property
  def occupancy_rate(self):
    """
    Get rate of schedules
    :return: float
    """
    return self._occupancy_rate

  @occupancy_rate.setter
  def occupancy_rate(self, value):
    """
    Set rate of schedules
    :param value: float
    """
    self._occupancy_rate = float(value)

  @property
  def occupant_type(self):
    """
    Get type of schedules
    :return: str
    """
    return self._occupant_type

  @occupant_type.setter
  def occupant_type(self, value):
    """
    Set type of schedules
    :param value: float
    """
    self._occupant_type = float(value)

  @property
  def arrival_time(self):
    """
    Get the arrival time of the occupant (for office building) in UTC with format YYYYMMDD HH:mm:ss
    :return: time
    """
    return self._arrival_time

  @arrival_time.setter
  def arrival_time(self, value):
    """
    Set the arrival time of the occupant (for office building) in UTC with format YYYYMMDD HH:mm:ss
    :param value: time
    """
    self._arrival_time = value

  @property
  def departure_time(self):
    """
    Get the departure time of the occupant (for office building) in UTC with format YYYYMMDD HH:mm:ss
    :return: time
    """
    return self._departure_time

  @departure_time.setter
  def departure_time(self, value):
    """
    Set the departure time of the occupant (for office building) in UTC with format YYYYMMDD HH:mm:ss
    :param value: str
    """
    self._departure_time = value

  @property
  def break_time(self):
    """
    Get the lunch or break time of the occupant (for office building) in UTC with format ????
    :return: break time
    """
    # todo @Sanam: define this format, is it the starting time? is it a list with both, starting and ending time?
    return self._break_time

  @property
  def day_of_week(self):
    """
    Get the day of the week (MON, TUE, WED, THU, FRI, SAT, SUN)
    :return: str
    """
    # todo @Sanam: is this a property or should it be a function
    #  to get the day of the week of an specific day of the year?
    return self._day_of_week

  @property
  def pd_of_meetings_duration(self):
    """
    Get the probability distribution of the meeting duration
    :return: ??
    """
    # todo @Sanam: what format are you expecting here??
    return self._pd_of_meetings_duration

  @pd_of_meetings_duration.setter
  def pd_of_meetings_duration(self, value):
    """
    Get the probability distribution of the meeting duration
    :param value: ??
    :return:
    """
    # todo @Sanam: what format are you expecting here??
    self._pd_of_meetings_duration = value



"""
LCA Material module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder atiya.atiya@mail.concordia.ca
"""

from typing import Union

class LcaMaterial:
  def __init__(self):
    self._id = None
    self._type = None
    self._name = None
    self._density = None
    self._density_unit = None
    self._embodied_carbon = None
    self._embodied_carbon_unit = None
    self._recycling_ratio = None
    self._company_recycling_ratio = None
    self._onsite_recycling_ratio = None
    self._landfilling_ratio = None
    self._cost = None
    self._cost_unit = None

  @property
  def id(self):
    """
    Get material id
    :return: int
    """
    return self._id

  @id.setter
  def id(self, value):
    """
    Set material id
    :param value: int
    """
    self._id = int(value)

  @property
  def type(self):
    """
    Get material type
    :return: str
    """
    return self._type

  @type.setter
  def type(self, value):
    """
    Set material type
    :param value: string
    """
    self._type = str(value)

  @property
  def name(self):
    """
    Get material name
    :return: str
    """
    return self._name

  @name.setter
  def name(self, value):
    """
    Set material name
    :param value: string
    """
    self._name = str(value)

  @property
  def density(self) -> Union[None, float]:
    """
    Get material density in kg/m3
    :return: None or float
    """
    return self._density

  @density.setter
  def density(self, value):
    """
    Set material density
    :param value: float
    """
    if value is not None:
      self._density = float(value)

  @property
  def density_unit(self) -> Union[None, str]:
    """
    Get material density unit
    :return: None or string
    """
    return self._density_unit

  @density_unit.setter
  def density_unit(self, value):
    """
    Set material density unit
    :param value: string
    """
    if value is not None:
      self._density_unit = str(value)

  @property
  def embodied_carbon(self) -> Union[None, float]:
    """
    Get material embodied carbon
    :return: None or float
    """
    return self._embodied_carbon

  @embodied_carbon.setter
  def embodied_carbon(self, value):
    """
    Set material embodied carbon
    :param value: float
    """
    if value is not None:
      self._embodied_carbon = float(value)

  @property
  def embodied_carbon_unit(self) -> Union[None, str]:
    """
    Get material embodied carbon unit
    :return: None or string
    """
    return self._embodied_carbon

  @embodied_carbon_unit.setter
  def embodied_carbon_unit(self, value):
    """
    Set material embodied carbon unit
    :param value: string
    """
    if value is not None:
      self._embodied_carbon_unit = str(value)

  @property
  def recycling_ratio(self) -> Union[None, float]:
    """
    Get material recycling ratio
    :return: None or float
    """
    return self._recycling_ratio

  @recycling_ratio.setter
  def recycling_ratio(self, value):
    """
    Set material recycling ratio
    :param value: float
    """
    if value is not None:
      self._recycling_ratio = float(value)

  @property
  def onsite_recycling_ratio(self) -> Union[None, float]:
    """
    Get material onsite recycling ratio
    :return: None or float
    """
    return self._onsite_recycling_ratio

  @onsite_recycling_ratio.setter
  def onsite_recycling_ratio(self, value):
    """
    Set material onsite recycling ratio
    :param value: float
    """
    if value is not None:
      self._onsite_recycling_ratio = float(value)

  @property
  def company_recycling_ratio(self) -> Union[None, float]:
    """
    Get material company recycling ratio
    :return: None or float
    """
    return self._company_recycling_ratio

  @company_recycling_ratio.setter
  def company_recycling_ratio(self, value):
    """
    Set material company recycling ratio
    :param value: float
    """
    if value is not None:
      self._company_recycling_ratio = float(value)

  @property
  def landfilling_ratio(self) -> Union[None, float]:
    """
    Get material landfilling ratio
    :return: None or float
    """
    return self._landfilling_ratio

  @landfilling_ratio.setter
  def landfilling_ratio(self, value):
    """
    Set material landfilling ratio
    :param value: float
    """
    if value is not None:
      self._landfilling_ratio = float(value)

  @property
  def cost(self) -> Union[None, float]:
    """
    Get material cost
    :return: None or float
    """
    return self._cost

  @cost.setter
  def cost(self, value):
    """
    Set material cost
    :param value: float
    """
    if value is not None:
      self._cost = float(value)

  @property
  def cost_unit(self) -> Union[None, str]:
    """
    Get material cost unit
    :return: None or string
    """
    return self._cost_unit

  @cost_unit.setter
  def cost_unit(self, value):
    """
    Set material cost unit
    :param value: string
    """
    if value is not None:
      self._cost_unit = float(value)



from setuptools import setup, find_packages, convert_path
import os.path
import glob

main_ns = {}
version = convert_path('hub/version.py')
with open(version) as f:
  exec(f.read(), main_ns)

setup(
  name='cerc-hub',
  version=main_ns['__version__'],
  description="CERC Hub consist in a set of classes (Central data model), importers and exporters to help researchers "
              "to create better and sustainable cities",
  long_description="CERC Hub consist in a set of classes (Central data model), importers and exporters to help "
                   "researchers to create better and sustainable cities.\n\nDevelop at Concordia university in canada "
                   "as part of the research group from the next generation cities institute our aim among others it's "
                   "to provide a comprehensive set of tools to help researchers and urban developers to make decisions "
                   "to improve the livability and efficiency of our cities",
  classifiers=[
    "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
  ],
  include_package_data=True,
  packages=['hub',
            'hub.catalog_factories',
            'hub.catalog_factories.construction',
            'hub.catalog_factories.data_models',
            'hub.catalog_factories.data_models.construction',
            'hub.catalog_factories.data_models.greenery',
            'hub.catalog_factories.data_models.usages',
            'hub.catalog_factories.greenery',
            'hub.catalog_factories.greenery.ecore_greenery',
            'hub.catalog_factories.usage',
            'hub.city_model_structure',
            'hub.city_model_structure.attributes',
            'hub.city_model_structure.building_demand',
            'hub.city_model_structure.energy_systems',
            'hub.city_model_structure.greenery',
            'hub.city_model_structure.iot',
            'hub.city_model_structure.transport',
            'hub.config',
            'hub.data',
            'hub.exports',
            'hub.exports.building_energy',
            'hub.exports.building_energy.idf_files',
            'hub.exports.building_energy.insel',
            'hub.exports.energy_systems',
            'hub.exports.formats',
            'hub.helpers',
            'hub.hub_logger',
            'hub.imports',
            'hub.imports.construction',
            'hub.imports.construction.helpers',
            'hub.imports.construction.data_classes',
            'hub.imports.energy_systems',
            'hub.imports.geometry',
            'hub.imports.geometry.citygml_classes',
            'hub.imports.geometry.helpers',
            'hub.imports.usage',
            'hub.imports.weather',
            'hub.imports.weather.helpers',
            'hub.persistence',
            'hub.persistence.models',
            'hub.persistence.repositories',
            'hub.imports'
            ],
  data_files=[
    ('hub/config', glob.glob('hub/config/*.ini')),
    ('hub/catalog_factories/greenery/ecore_greenery', glob.glob('hub/catalog_factories/greenery/ecore_greenery/*.ecore')),
    ('hub/data/construction.', glob.glob('hub/data/construction/*.xml')),
    ('hub/data/customized_imports/', glob.glob('hub/data/customized_imports/*.xml')),
    ('hub/data/energy_systems/', glob.glob('hub/data/energy_systems/*.xml')),
    ('hub/data/energy_systems/', glob.glob('hub/data/energy_systems/*.insel')),
    ('hub/data/energy_systems/', glob.glob('hub/data/energy_systems/*.xlsx')),
    ('hub/data/energy_systems/*', glob.glob('hub/data/energy_systems/*.txt')),
    ('hub/data/energy_systems/*', glob.glob('hub/data/energy_systems/*.yaml')),
    ('hub/data/greenery/', glob.glob('hub/data/greenery/*.xml')),
    ('hub/data/life_cycle_assessment/', glob.glob('hub/data/life_cycle_assessment/*.xml')),
    ('hub/data/schedules/', glob.glob('hub/data/schedules/*.xml')),
    ('hub/data/schedules/', glob.glob('hub/data/schedules/*.xlsx')),
    ('hub/data/schedules/idf_files/', glob.glob('hub/data/schedules/idf_files/*.idf')),
    ('hub/data/sensors/', glob.glob('hub/data/sensors/*.json')),
    ('hub/data/usage/', glob.glob('hub/data/usage/*.xml')),
    ('hub/data/usage/', glob.glob('hub/data/usage/*.xlsx')),
    ('hub/data/weather/', glob.glob('hub/data/weather/*.dat')),
    ('hub/data/weather/epw/', glob.glob('hub/data/weather/epw/*.epw')),
    ('hub/data/weather/', glob.glob('hub/data/weather/*.dat')),
    ('hub/exports/building_energy/idf_files', glob.glob('hub/exports/building_energy/idf_files/*.idf'))
  ],
  setup_requires=['setuptools']
)



"""
Construction catalog window
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""


class Window:
  def __init__(self, window_id, frame_ratio, g_value, overall_u_value, name):
    self._id = window_id
    self._frame_ratio = frame_ratio
    self._g_value = g_value
    self._overall_u_value = overall_u_value
    self._name = name

  @property
  def id(self):
    """
    Get window id
    :return: str
    """
    return self._id

  @property
  def name(self):
    """
    Get window name
    :return: str
    """
    return self._name

  @property
  def frame_ratio(self):
    """
    Get window frame ratio
    :return: float
    """
    return self._frame_ratio

  @property
  def g_value(self):
    """
    Get thermal opening g-value
    :return: float
    """
    return self._g_value

  @property
  def overall_u_value(self):
    """
    Get thermal opening overall U-value in W/m2K
    :return: float
    """
    return self._overall_u_value



"""
Configuration helper
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""
import configparser
from pathlib import Path


class ConfigurationHelper:
  """
  Configuration class
  """
  def __init__(self):
    config_file = Path(Path(__file__).parent.parent / 'config/configuration.ini').resolve()
    self._config = configparser.ConfigParser()
    self._config.read(config_file)

  @property
  def max_location_distance_for_shared_walls(self) -> float:
    """
    Get configured maximal distance between attributes to consider that they may share walls in meters
    :return: 5.0
    """
    return self._config.getfloat('buildings', 'max_location_distance_for_shared_walls')

  @property
  def min_coordinate(self) -> float:
    """
    Get configured minimal coordinate value
    :return: -1.7976931348623157e+308
    """
    return self._config.getfloat('buildings', 'min_coordinate')

  @property
  def max_coordinate(self) -> float:
    """
    Get configured maximal coordinate value
    :return: 1.7976931348623157e+308
    """
    return self._config.getfloat('buildings', 'max_coordinate').real

  @property
  def comnet_lighting_latent(self) -> float:
    """
    Get configured latent ratio of internal gains do to lighting used for Comnet (ASHRAE) standard
    :return: 0
    """
    return self._config.getfloat('buildings', 'comnet_lighting_latent').real

  @property
  def comnet_lighting_convective(self) -> float:
    """
    Get configured convective ratio of internal gains do to lighting used for Comnet (ASHRAE) standard
    :return: 0.5
    """
    return self._config.getfloat('buildings', 'comnet_lighting_convective').real

  @property
  def comnet_lighting_radiant(self) -> float:
    """
    Get configured radiant ratio of internal gains do to lighting used for Comnet (ASHRAE) standard
    :return: 0.5
    """
    return self._config.getfloat('buildings', 'comnet_lighting_radiant').real

  @property
  def comnet_plugs_latent(self) -> float:
    """
    Get configured latent ratio of internal gains do to electrical appliances used for Comnet (ASHRAE) standard
    :return: 0
    """
    return self._config.getfloat('buildings', 'comnet_plugs_latent').real

  @property
  def comnet_plugs_convective(self) -> float:
    """
    Get configured convective ratio of internal gains do to electrical appliances used for Comnet (ASHRAE) standard
    :return: 0.75
    """
    return self._config.getfloat('buildings', 'comnet_plugs_convective').real

  @property
  def comnet_plugs_radiant(self) -> float:
    """
    Get configured radiant ratio of internal gains do to electrical appliances used for Comnet (ASHRAE) standard
    :return: 0.25
    """
    return self._config.getfloat('buildings', 'comnet_plugs_radiant').real

  @property
  def comnet_occupancy_sensible_convective(self) -> float:
    """
    Get configured convective ratio of the sensible part of internal gains do to occupancy
    used for Comnet (ASHRAE) standard
    :return: 0.9
    """
    return self._config.getfloat('buildings', 'comnet_occupancy_sensible_convective').real

  @property
  def comnet_occupancy_sensible_radiant(self) -> float:
    """
    Get configured radiant ratio of the sensible part of internal gains do to occupancy
    used for Comnet (ASHRAE) standard
    :return: 0.1
    """
    return self._config.getfloat('buildings', 'comnet_occupancy_sensible_radiant').real

  @property
  def convective_heat_transfer_coefficient_interior(self) -> float:
    """
    Get configured convective heat transfer coefficient for surfaces inside the building
    :return: 3.5 W/m2K
    """
    return self._config.getfloat('buildings', 'convective_heat_transfer_coefficient_interior').real

  @property
  def convective_heat_transfer_coefficient_exterior(self) -> float:
    """
    Get configured convective heat transfer coefficient for surfaces outside the building
    :return: 20 W/m2K
    """
    return self._config.getfloat('buildings', 'convective_heat_transfer_coefficient_exterior').real

  @property
  def soil_conductivity(self) -> float:
    """
    Get configured soil conductivity for surfaces touching the ground
    :return: 3 W/mK
    """
    return self._config.getfloat('buildings', 'soil_conductivity').real

  @property
  def soil_thickness(self) -> float:
    """
    Get configured soil thickness for surfaces touching the ground
    :return: 0.5
    """
    return self._config.getfloat('buildings', 'soil_thickness').real



"""
weather helper
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
import math
import hub.helpers.constants as cte


class Weather:
  """
  Weather class
  """

  @staticmethod
  def sky_temperature(ambient_temperature):
    """
    Get sky temperature from ambient temperature in Celsius
    :return: float
    """
    # Swinbank - Source sky model approximation(1963) based on cloudiness statistics(32 %) in United States
    # ambient temperatures( in °C)
    # sky temperatures( in °C)
    values = []
    for temperature in ambient_temperature:
      value = 0.037536 * math.pow((temperature + cte.KELVIN), 1.5) \
              + 0.32 * (temperature + cte.KELVIN) - cte.KELVIN
      values.append(value)
    return values



"""
CityGml module parses citygml_classes files and import the geometry into the city model structure
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Atiya atiya.atiya@mail.concordia.ca
"""
import xmltodict
from pathlib import Path
from hub.city_model_structure.vehicle import Vehicle


class LcaVehicle:
  def __init__(self, city, base_path):
    self._city = city
    self._base_path = base_path
    self._lca = None

  def enrich(self):
    self._city.vehicles = []
    path = Path(self._base_path / 'lca_data.xml').resolve()

    with open(path) as xml:
      self._lca = xmltodict.parse(xml.read())
    for vehicle in self._lca["library"]["vehicles"]['vehicle']:
      self._city.vehicles.append(Vehicle(vehicle['@id'], vehicle['@name'], vehicle['fuel_consumption_rate']['#text'],
                                         vehicle['fuel_consumption_rate']['@unit'],
                                         vehicle['carbon_emission_factor']['#text'],
                                         vehicle['carbon_emission_factor']['@unit']))



"""
Model representation of simulation results
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""

import datetime

from sqlalchemy import Column, Integer, String, Sequence, ForeignKey
from sqlalchemy import DateTime
from sqlalchemy.dialects.postgresql import JSONB
from hub.persistence.configuration import Models

class SimulationResults(Models):
  """
  A model representation of an application
  """
  __tablename__ = 'simulation_results'
  id = Column(Integer, Sequence('application_id_seq'), primary_key=True)
  city_id = Column(Integer, ForeignKey('city.id'), nullable=True)
  city_object_id = Column(Integer, ForeignKey('city_object.id'), nullable=True)
  name = Column(String, nullable=False)
  values = Column(JSONB, nullable=False)
  created = Column(DateTime, default=datetime.datetime.utcnow)
  updated = Column(DateTime, default=datetime.datetime.utcnow)

  def __init__(self, name, values, city_id=None, city_object_id=None):
    self.name = name
    self.values = values
    self.city_id = city_id
    self.city_object_id = city_object_id



"""
HvacSystem module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from typing import Union, List
from hub.city_model_structure.building_demand.thermal_zone import ThermalZone


class HvacSystem:
  """
  HvacSystem class
  """
  def __init__(self):
    self._type = None
    self._thermal_zones = None

  @property
  def type(self) -> Union[None, str]:
    """
    Get hvac system type
    :return: None or str
    """
    return self._type

  @type.setter
  def type(self, value):
    """
    Set hvac system type
    :param value: str
    """
    if value is not None:
      self._type = str(value)

  @property
  def thermal_zones(self) -> Union[None, List[ThermalZone]]:
    """
    Get list of zones that this unit serves
    :return: None or [ThermalZone]
    """
    return self._thermal_zones

  @thermal_zones.setter
  def thermal_zones(self, value):
    """
    Set list of zones that this unit serves
    :param value: [ThermalZone]
    """
    self._thermal_zones = value



"""
Greenery catalog
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from pyecore.resources import ResourceSet, URI
from hub.catalog_factories.greenery.ecore_greenery.greenerycatalog import GreeneryCatalog as gc
from hub.catalog_factories.catalog import Catalog
from pathlib import Path
from hub.catalog_factories.data_models.greenery.vegetation import Vegetation as libs_vegetation
from hub.catalog_factories.data_models.greenery.plant import Plant as libs_plant
from hub.catalog_factories.data_models.greenery.soil import Soil as libs_soil
from hub.catalog_factories.data_models.greenery.plant_percentage import PlantPercentage as libs_pp
from hub.catalog_factories.data_models.greenery.content import Content as GreeneryContent


class GreeneryCatalog(Catalog):

  def __init__(self, path):
    base_path = Path(Path(__file__).parent / 'ecore_greenery' / 'greenerycatalog_no_quantities.ecore')
    resource_set = ResourceSet()
    data_model = resource_set.get_resource(URI(str(base_path)))
    data_model_root = data_model.contents[0]
    resource_set.metamodel_registry[data_model_root.nsURI] = data_model_root
    resource = resource_set.get_resource(URI(str(path)))
    catalog_data: gc = resource.contents[0]

    plants = []
    for plant_category in catalog_data.plantCategories:
      name = plant_category.name
      for plant in plant_category.plants:
        plants.append(libs_plant(name, plant))

    vegetations = []
    for vegetation_category in catalog_data.vegetationCategories:
      name = vegetation_category.name
      for vegetation in vegetation_category.vegetationTemplates:
        plant_percentages = []

        for plant_percentage in vegetation.plants:
          plant_category = "Unknown"
          for plant in plants:
            if plant.name == plant_percentage.plant.name:
              plant_category = plant.category
              break
          plant_percentages.append(libs_pp(plant_percentage.percentage,plant_category, plant_percentage.plant))
        vegetations.append(libs_vegetation(name, vegetation, plant_percentages))
    plants = []
    for plant_category in catalog_data.plantCategories:
      name = plant_category.name
      for plant in plant_category.plants:
        plants.append(libs_plant(name, plant))

    soils = []
    for soil in catalog_data.soils:
      soils.append(libs_soil(soil))

    self._content = GreeneryContent(vegetations, plants, soils)

  def names(self, category=None):
    """
    Get the catalog elements names
    :parm: optional category filter
    """
    if category is None:
      _names = {'vegetations': [], 'plants': [], 'soils': []}
      for vegetation in self._content.vegetations:
        _names['vegetations'].append(vegetation.name)
      for plant in self._content.plants:
        _names['plants'].append(plant.name)
      for soil in self._content.soils:
        _names['soils'].append(soil.name)
    else:
      _names = {category: []}
      if category.lower() == 'vegetations':
        for vegetation in self._content.vegetations:
          _names[category].append(vegetation.name)
      elif category.lower() == 'plants':
        for plant in self._content.plants:
          _names[category].append(plant.name)
      elif category.lower() == 'soils':
        for soil in self._content.soils:
          _names[category].append(soil.name)
      else:
        raise ValueError(f'Unknown category [{category}]')
    return _names

  def get_entry(self, name):
    """
    Get one complete entry from the greenery catalog
    """
    for entry in self._content.vegetations:
      if entry.name.lower() == name.lower():
        return entry
    for entry in self._content.plants:
      if entry.name.lower() == name.lower():
        return entry
    for entry in self._content.soils:
      if entry.name.lower() == name.lower():
        return entry
    raise IndexError(f"{name} doesn't exists in the catalog")

  def entries(self, category=None):
    if category is None:
      return self._content
    else:
      if category.lower() == 'vegetations':
        return self._content.vegetations
      elif category.lower() == 'plants':
        return self._content.plants
      elif category.lower() == 'soils':
        return self._content.soils
      else:
        raise ValueError(f'Unknown category [{category}]')



"""
Polygon module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from __future__ import annotations
import math
import sys
from typing import List
import numpy as np
from trimesh import Trimesh
import trimesh.intersections

from hub.city_model_structure.attributes.plane import Plane
from hub.city_model_structure.attributes.point import Point
import hub.helpers.constants as cte


class Polygon:
  """
  Polygon class
  """

  def __init__(self, coordinates):
    self._area = None
    self._points = None
    self._points_list = None
    self._normal = None
    self._inverse = None
    self._edges = None
    self._coordinates = coordinates
    self._triangles = None
    self._vertices = None
    self._faces = None
    self._plane = None

  @property
  def points(self) -> List[Point]:
    """
    Get the points belonging to the polygon [[x, y, z],...]
    :return: [Point]
    """
    if self._points is None:
      self._points = []
      for coordinate in self.coordinates:
        self._points.append(Point(coordinate))
    return self._points

  @property
  def plane(self) -> Plane:
    """
    Get the polygon plane
    :return: Plane
    """
    if self._plane is None:
      self._plane = Plane(normal=self.normal, origin=self.points[0])
    return self._plane

  @property
  def coordinates(self) -> List[np.ndarray]:
    """
    Get the points in the shape of its coordinates belonging to the polygon [[x, y, z],...]
    :return: [np.ndarray]
    """
    return self._coordinates

  @staticmethod
  def _module(vector):
    x2 = vector[0] ** 2
    y2 = vector[1] ** 2
    z2 = vector[2] ** 2
    return math.sqrt(x2+y2+z2)

  @staticmethod
  def _scalar_product(vector_0, vector_1):
    x = vector_0[0] * vector_1[0]
    y = vector_0[1] * vector_1[1]
    z = vector_0[2] * vector_1[2]
    return x+y+z

  def contains_point(self, point):
    """
    Determines if the given point is contained by the current polygon
    :return: boolean
    """
    # fixme: This method doesn't seems to work.
    n = len(self.vertices)
    angle_sum = 0
    for i in range(0, n):
      vector_0 = self.vertices[i]
      vector_1 = self.vertices[(i+1) % n]
      # set to origin
      vector_0[0] = vector_0[0] - point.coordinates[0]
      vector_0[1] = vector_0[1] - point.coordinates[1]
      vector_0[2] = vector_0[2] - point.coordinates[2]
      vector_1[0] = vector_1[0] - point.coordinates[0]
      vector_1[1] = vector_1[1] - point.coordinates[1]
      vector_1[2] = vector_1[2] - point.coordinates[2]
      module = Polygon._module(vector_0) * Polygon._module(vector_1)

      scalar_product = Polygon._scalar_product(vector_0, vector_1)
      angle = np.pi/2
      if module != 0:
        angle = abs(np.arcsin(scalar_product / module))
      angle_sum += angle
    return abs(angle_sum - math.pi*2) < cte.EPSILON

  def contains_polygon(self, polygon):
    """
    Determines if the given polygon is contained by the current polygon
    :return: boolean
    """

    for point in polygon.points:
      if not self.contains_point(point):
        return False
    return True

  @property
  def points_list(self) -> np.ndarray:
    """
    Get the solid surface point coordinates list [x, y, z, x, y, z,...]
    :return: np.ndarray
    """
    if self._points_list is None:
      s = self.coordinates
      self._points_list = np.reshape(s, len(s) * 3)
    return self._points_list

  @property
  def edges(self) -> List[List[Point]]:
    """
    Get polygon edges list
    :return: [[Point]]
    """
    if self._edges is None:
      self._edges = []
      for i in range(0, len(self.points) - 1):
        point_1 = self.points[i]
        point_2 = self.points[i + 1]
        self._edges.append([point_1, point_2])
      self._edges.append([self.points[len(self.points) - 1], self.points[0]])
    return self._edges

  @property
  def area(self):
    """
    Get surface area in square meters
    :return: float
    """
    # New method to calculate area
    if self._area is None:
      if len(self.points) < 3:
        sys.stderr.write('Warning: the area of a line or point cannot be calculated 1. Area = 0\n')
        return 0
      alpha = 0
      vec_1 = self.points[1].coordinates - self.points[0].coordinates
      for i in range(2, len(self.points)):
        vec_2 = self.points[i].coordinates - self.points[0].coordinates
        alpha += self._angle_between_vectors(vec_1, vec_2)
      if alpha == 0:
        sys.stderr.write('Warning: the area of a line or point cannot be calculated 2. Area = 0\n')
        return 0
      horizontal_points = self._points_rotated_to_horizontal
      area = 0
      for i in range(0, len(horizontal_points) - 1):
        point = horizontal_points[i]
        next_point = horizontal_points[i + 1]
        area += (next_point[1] + point[1]) / 2 * (next_point[0] - point[0])
      next_point = horizontal_points[0]
      point = horizontal_points[len(horizontal_points) - 1]
      area += (next_point[1] + point[1]) / 2 * (next_point[0] - point[0])
      self._area = abs(area)
    return self._area

  @property
  def _points_rotated_to_horizontal(self):
    """
    polygon points rotated to horizontal
    :return: [float]
    """
    z_vector = [0, 0, 1]
    normal_vector = self.normal
    horizontal_points = []
    x = normal_vector[0]
    y = normal_vector[1]

    if x == 0 and y == 0:
      # Already horizontal
      for point in self.points:
        horizontal_points.append([point.coordinates[0], point.coordinates[1], 0])
    else:
      alpha = self._angle_between_vectors(normal_vector, z_vector)
      rotation_line = np.cross(normal_vector, z_vector)
      third_axis = np.cross(normal_vector, rotation_line)
      w_1 = rotation_line / np.linalg.norm(rotation_line)
      w_2 = normal_vector
      w_3 = third_axis / np.linalg.norm(third_axis)
      rotation_matrix = np.array([[1, 0, 0],
                                  [0, np.cos(alpha), -np.sin(alpha)],
                                  [0, np.sin(alpha), np.cos(alpha)]])
      base_matrix = np.array([w_1, w_2, w_3])
      rotation_base_matrix = np.matmul(base_matrix.transpose(), rotation_matrix.transpose())
      rotation_base_matrix = np.matmul(rotation_base_matrix, base_matrix)

      if rotation_base_matrix is None:
        sys.stderr.write('Warning: rotation base matrix returned None\n')
      else:
        for point in self.points:
          new_point = np.matmul(rotation_base_matrix, point.coordinates)
          horizontal_points.append(new_point)
    return horizontal_points

  @property
  def normal(self) -> np.ndarray:
    """
    Get surface normal vector
    :return: np.ndarray
    """
    if self._normal is None:
      points = self.coordinates
      # todo: IF THE FIRST ONE IS 0, START WITH THE NEXT
      point_origin = points[len(points) - 2]
      vector_1 = points[len(points) - 1] - point_origin
      vector_2 = points[0] - point_origin
      vector_3 = points[1] - point_origin
      cross_product = np.cross(vector_1, vector_2)
      if np.linalg.norm(cross_product) != 0:
        cross_product = cross_product / np.linalg.norm(cross_product)
        alpha = self._angle_between_vectors(vector_1, vector_2)
      else:
        cross_product = [0, 0, 0]
        alpha = 0
      if len(points) == 3:
        return cross_product
      if np.linalg.norm(cross_product) == 0:
        return cross_product
      alpha += self._angle(vector_2, vector_3, cross_product)
      for i in range(0, len(points) - 4):
        vector_1 = points[i + 1] - point_origin
        vector_2 = points[i + 2] - point_origin
        alpha += self._angle(vector_1, vector_2, cross_product)
      vector_1 = points[len(points) - 1] - point_origin
      vector_2 = points[0] - point_origin
      if alpha < 0:
        cross_product = np.cross(vector_2, vector_1)
      else:
        cross_product = np.cross(vector_1, vector_2)
      self._normal = cross_product / np.linalg.norm(cross_product)
    return self._normal

  @staticmethod
  def _angle(vector_1, vector_2, cross_product):
    """
    alpha angle in radians
    :param vector_1: [float]
    :param vector_2: [float]
    :param cross_product: [float]
    :return: float
    """
    accepted_normal_difference = 0.01
    cross_product_next = np.cross(vector_1, vector_2)
    if np.linalg.norm(cross_product_next) != 0:
      cross_product_next = cross_product_next / np.linalg.norm(cross_product_next)
      alpha = Polygon._angle_between_vectors(vector_1, vector_2)
    else:
      cross_product_next = [0, 0, 0]
      alpha = 0
    delta_normals = 0
    for j in range(0, 3):
      delta_normals += cross_product[j] - cross_product_next[j]
    if np.abs(delta_normals) < accepted_normal_difference:
      return alpha
    return -alpha

  def triangulate(self) -> List[Polygon]:
    """
    Triangulates a polygon following the ear clipping methodology
    :return: list[triangles]
    """
    # todo: review triangulate_polygon in
    #  https://github.com/mikedh/trimesh/blob/dad11126742e140ef46ba12f8cb8643c83356467/trimesh/creation.py#L415,
    #  it had a problem with a class called 'triangle', but, if solved,
    #  it could be a very good substitute of this method
    #  this method is very dirty and has an infinite loop solved with a counter!!
    if self._triangles is None:
      points_list = self.points_list
      normal = self.normal
      if np.linalg.norm(normal) == 0:
        sys.stderr.write('Not able to triangulate polygon\n')
        return [self]
      # are points concave or convex?
      total_points_list, concave_points, convex_points = self._starting_lists(points_list, normal)

      # list of ears
      ears = []
      j = 0
      while (len(concave_points) > 3 or len(convex_points) != 0) and j < 100:
        j += 1
        for i in range(0, len(concave_points)):
          ear = self._triangle(points_list, total_points_list, concave_points[i])
          rest_points = []
          for points in total_points_list:
            rest_points.append(list(self.coordinates[points]))
          if self._is_ear(ear, rest_points):
            ears.append(ear)
            point_to_remove = concave_points[i]
            previous_point_in_list, next_point_in_list = self._enveloping_points(point_to_remove,
                                                                                 total_points_list)
            total_points_list.remove(point_to_remove)
            concave_points.remove(point_to_remove)
            # Was any of the adjacent points convex? -> check if changed status to concave
            for convex_point in convex_points:
              if convex_point == previous_point_in_list:
                concave_points, convex_points, end_loop = self._if_concave_change_status(normal,
                                                                                         points_list,
                                                                                         convex_point,
                                                                                         total_points_list,
                                                                                         concave_points,
                                                                                         convex_points,
                                                                                         previous_point_in_list)
                if end_loop:
                  break
                continue
              if convex_point == next_point_in_list:
                concave_points, convex_points, end_loop = self._if_concave_change_status(normal,
                                                                                         points_list,
                                                                                         convex_point,
                                                                                         total_points_list,
                                                                                         concave_points,
                                                                                         convex_points,
                                                                                         next_point_in_list)
                if end_loop:
                  break
                continue
            break
        if len(total_points_list) <= 3 and len(convex_points) > 0:
          sys.stderr.write('Not able to triangulate polygon\n')
          return [self]
      if j >= 100:
        sys.stderr.write('Not able to triangulate polygon\n')
        return [self]
      last_ear = self._triangle(points_list, total_points_list, concave_points[1])
      ears.append(last_ear)
      self._triangles = ears
    return self._triangles

  @staticmethod
  def _starting_lists(points_list, normal) -> [List[float], List[float], List[float]]:
    """
    creates the list of vertices (points) that define the polygon (total_points_list), together with other two lists
    separating points between convex and concave
    :param points_list: points_list
    :param normal: normal
    :return: list[point], list[point], list[point]
    """
    concave_points = []
    convex_points = []
    # lists of concave and convex points
    # case 1: first point
    point = points_list[0:3]
    previous_point = points_list[len(points_list) - 3:]
    next_point = points_list[3:6]
    index = 0
    total_points_list = [index]
    if Polygon._point_is_concave(normal, point, previous_point, next_point):
      concave_points.append(index)
    else:
      convex_points.append(index)
    # case 2: all points except first and last
    for i in range(0, int((len(points_list) - 6) / 3)):
      point = points_list[(i + 1) * 3:(i + 2) * 3]
      previous_point = points_list[i * 3:(i + 1) * 3]
      next_point = points_list[(i + 2) * 3:(i + 3) * 3]
      index = i + 1
      total_points_list.append(index)
      if Polygon._point_is_concave(normal, point, previous_point, next_point):
        concave_points.append(index)
      else:
        convex_points.append(index)
    # case 3: last point
    point = points_list[len(points_list) - 3:]
    previous_point = points_list[len(points_list) - 6:len(points_list) - 3]
    next_point = points_list[0:3]
    index = int(len(points_list) / 3) - 1
    total_points_list.append(index)
    if Polygon._point_is_concave(normal, point, previous_point, next_point):
      concave_points.append(index)
    else:
      convex_points.append(index)
    return total_points_list, concave_points, convex_points

  @staticmethod
  def _triangle(points_list, total_points_list, point_position) -> Polygon:
    """
    creates a triangular polygon out of three points
    :param points_list: points_list
    :param total_points_list: [point]
    :param point_position: int
    :return: polygon
    """
    index = point_position * 3
    previous_point_index, next_point_index = Polygon._enveloping_points_indices(point_position, total_points_list)
    points = points_list[previous_point_index:previous_point_index + 3]
    points = np.append(points, points_list[index:index + 3])
    points = np.append(points, points_list[next_point_index:next_point_index + 3])
    rows = points.size // 3
    points = points.reshape(rows, 3)
    triangle = Polygon(points)
    return triangle

  @staticmethod
  def _enveloping_points_indices(point_position, total_points_list):
    """
    due to the fact that the lists are not circular, a method to find the previous and next points
    of an specific one is needed
    :param point_position: int
    :param total_points_list: [point]
    :return: int, int
    """
    previous_point_index = None
    next_point_index = None
    if point_position == total_points_list[0]:
      previous_point_index = total_points_list[len(total_points_list) - 1] * 3
      next_point_index = total_points_list[1] * 3
    if point_position == total_points_list[len(total_points_list) - 1]:
      previous_point_index = total_points_list[len(total_points_list) - 2] * 3
      next_point_index = total_points_list[0] * 3
    for i in range(1, len(total_points_list) - 1):
      if point_position == total_points_list[i]:
        previous_point_index = total_points_list[i - 1] * 3
        next_point_index = total_points_list[i + 1] * 3
    return previous_point_index, next_point_index

  @staticmethod
  def _enveloping_points(point_to_remove, total_points_list):
    """
    due to the fact that the lists are not circular, a method to find the previous and next points
    of an specific one is needed
    :param point_to_remove: point
    :param total_points_list: [point]
    :return: point, point
    """
    index = total_points_list.index(point_to_remove)
    if index == 0:
      previous_point_in_list = total_points_list[len(total_points_list) - 1]
      next_point_in_list = total_points_list[1]
    elif index == len(total_points_list) - 1:
      previous_point_in_list = total_points_list[len(total_points_list) - 2]
      next_point_in_list = total_points_list[0]
    else:
      previous_point_in_list = total_points_list[index - 1]
      next_point_in_list = total_points_list[index + 1]
    return previous_point_in_list, next_point_in_list

  @staticmethod
  def _is_ear(ear, points) -> bool:
    """
    finds whether a triangle is an ear of the polygon
    :param ear: polygon
    :param points: [point]
    :return: boolean
    """
    area_ear = ear.area
    for point in points:
      area_points = 0
      point_is_not_vertex = True
      for i in range(0, 3):
        if abs(np.linalg.norm(point) - np.linalg.norm(ear.coordinates[i])) < 0.0001:
          point_is_not_vertex = False
          break
      if point_is_not_vertex:
        for i in range(0, 3):
          if i != 2:
            new_points = ear.coordinates[i][:]
            new_points = np.append(new_points, ear.coordinates[i + 1][:])
            new_points = np.append(new_points, point[:])
          else:
            new_points = ear.coordinates[i][:]
            new_points = np.append(new_points, point[:])
            new_points = np.append(new_points, ear.coordinates[0][:])
          rows = new_points.size // 3
          new_points = new_points.reshape(rows, 3)
          new_triangle = Polygon(new_points)
          area_points += new_triangle.area
      if abs(area_points - area_ear) < 1e-6:
        # point_inside_ear = True
        return False
    return True

  @staticmethod
  def _if_concave_change_status(normal, points_list, convex_point, total_points_list,
                                concave_points, convex_points, point_in_list) -> [List[float], List[float], bool]:
    """
    checks whether an convex specific point change its status to concave after removing one ear in the polygon
    returning the new convex and concave points lists together with a flag advising that the list of total points
    already 3 and, therefore, the triangulation must be finished.
    :param normal: normal
    :param points_list: points_list
    :param convex_point: int
    :param total_points_list: [point]
    :param concave_points: [point]
    :param convex_points: [point]
    :param point_in_list: int
    :return: list[points], list[points], boolean
    """
    end_loop = False
    point = points_list[point_in_list * 3:(point_in_list + 1) * 3]
    pointer = total_points_list.index(point_in_list) - 1
    if pointer < 0:
      pointer = len(total_points_list) - 1
    previous_point = points_list[total_points_list[pointer] * 3:total_points_list[pointer] * 3 + 3]
    pointer = total_points_list.index(point_in_list) + 1
    if pointer >= len(total_points_list):
      pointer = 0
    next_point = points_list[total_points_list[pointer] * 3:total_points_list[pointer] * 3 + 3]
    if Polygon._point_is_concave(normal, point, previous_point, next_point):
      if concave_points[0] > convex_point:
        concave_points.insert(0, convex_point)
      elif concave_points[len(concave_points) - 1] < convex_point:
        concave_points.append(convex_point)
      else:
        for point_index in range(0, len(concave_points) - 1):
          if concave_points[point_index] < convex_point < concave_points[point_index + 1]:
            concave_points.insert(point_index + 1, convex_point)
      convex_points.remove(convex_point)
      end_loop = True
    return concave_points, convex_points, end_loop

  @staticmethod
  def _point_is_concave(normal, point, previous_point, next_point) -> bool:
    """
    returns whether a point is concave
    :param normal: normal
    :param point: point
    :param previous_point: point
    :param next_point: point
    :return: boolean
    """
    is_concave = False
    accepted_error = 0.1
    points = np.append(previous_point, point)
    points = np.append(points, next_point)
    rows = points.size // 3
    points = points.reshape(rows, 3)
    triangle = Polygon(points)
    error_sum = 0
    for i in range(0, len(normal)):
      error_sum += triangle.normal[i] - normal[i]
    if np.abs(error_sum) < accepted_error:
      is_concave = True
    return is_concave

  @staticmethod
  def _angle_between_vectors(vec_1, vec_2):
    """
    angle between vectors in radians
    :param vec_1: vector
    :param vec_2: vector
    :return: float
    """
    if np.linalg.norm(vec_1) == 0 or np.linalg.norm(vec_2) == 0:
      sys.stderr.write("Warning: impossible to calculate angle between planes' normal. Return 0\n")
      return 0
    cosine = np.dot(vec_1, vec_2) / np.linalg.norm(vec_1) / np.linalg.norm(vec_2)
    if cosine > 1 and cosine - 1 < 1e-5:
      cosine = 1
    elif cosine < -1 and cosine + 1 > -1e-5:
      cosine = -1
    alpha = math.acos(cosine)
    return alpha

  @property
  def inverse(self):
    """
    Get the polygon coordinates in reversed order
    :return: [np.ndarray]
    """
    if self._inverse is None:
      self._inverse = self.coordinates[::-1]
    return self._inverse

  def divide(self, plane):
    """
    Divides the polygon in two by a plane
    :param plane: plane that intersects with self to divide it in two parts (Plane)
    :return: Polygon, Polygon, [Point]
    """
    tri_polygons = Trimesh(vertices=self.vertices, faces=self.faces)
    intersection = trimesh.intersections.mesh_plane(tri_polygons, plane.normal, plane.origin.coordinates)
    polys_1 = trimesh.intersections.slice_mesh_plane(tri_polygons, plane.opposite_normal, plane.origin.coordinates)
    polys_2 = trimesh.intersections.slice_mesh_plane(tri_polygons, plane.normal, plane.origin.coordinates)
    triangles_1 = []
    for triangle in polys_1.triangles:
      triangles_1.append(Polygon(triangle))
    polygon_1 = self._reshape(triangles_1)
    triangles_2 = []
    for triangle in polys_2.triangles:
      triangles_2.append(Polygon(triangle))
    polygon_2 = self._reshape(triangles_2)
    return polygon_1, polygon_2, intersection

  def _reshape(self, triangles) -> Polygon:
    edges_list = []
    for i in range(0, len(triangles)):
      for edge in triangles[i].edges:
        if not self._edge_in_edges_list(edge, edges_list):
          edges_list.append(edge)
        else:
          edges_list = self._remove_from_list(edge, edges_list)
    points = self._order_points(edges_list)
    return Polygon(points)

  @staticmethod
  def _edge_in_edges_list(edge, edges_list):
    for edge_element in edges_list:
      if (edge_element[0].distance_to_point(edge[0]) == 0 and edge_element[1].distance_to_point(edge[1]) == 0) or \
          (edge_element[1].distance_to_point(edge[0]) == 0 and edge_element[0].distance_to_point(
            edge[1]) == 0):
        return True
    return False

  @staticmethod
  def _order_points(edges_list):
    # todo: not sure that this method works for any case -> RECHECK
    points = edges_list[0]
    for _ in range(0, len(points)):
      for i in range(1, len(edges_list)):
        point_1 = edges_list[i][0]
        point_2 = points[len(points) - 1]
        if point_1.distance_to_point(point_2) == 0:
          points.append(edges_list[i][1])
    points.remove(points[len(points) - 1])
    array_points = []
    for point in points:
      array_points.append(point.coordinates)
    return np.array(array_points)

  @staticmethod
  def _remove_from_list(edge, edges_list):
    new_list = []
    for edge_element in edges_list:
      if not ((edge_element[0].distance_to_point(edge[0]) == 0 and edge_element[1].distance_to_point(
          edge[1]) == 0) or
              (edge_element[1].distance_to_point(edge[0]) == 0 and edge_element[0].distance_to_point(
                edge[1]) == 0)):
        new_list.append(edge_element)
    return new_list

  @property
  def vertices(self) -> np.ndarray:
    """
    Get polyhedron vertices
    :return: np.ndarray(int)
    """
    if self._vertices is None:
      vertices, self._vertices = [], []
      _ = [vertices.extend(s.coordinates) for s in self.triangulate()]
      for vertex_1 in vertices:
        found = False
        for vertex_2 in self._vertices:
          found = False
          power = 0
          for dimension in range(0, 3):
            power += math.pow(vertex_2[dimension] - vertex_1[dimension], 2)
          distance = math.sqrt(power)
          if distance == 0:
            found = True
            break
        if not found:
          self._vertices.append(vertex_1)
      self._vertices = np.asarray(self._vertices)
    return self._vertices

  @property
  def faces(self) -> List[List[int]]:
    """
    Get polyhedron triangular faces
    :return: [face]
    """
    if self._faces is None:
      self._faces = []

      for polygon in self.triangulate():
        face = []
        points = polygon.coordinates
        if len(points) != 3:
          sub_polygons = polygon.triangulate()
          # todo: I modified this! To be checked @Guille
          if len(sub_polygons) >= 1:
            for sub_polygon in sub_polygons:
              face = []
              points = sub_polygon.coordinates
              for point in points:
                face.append(self._position_of(point, face))
              self._faces.append(face)
        else:
          for point in points:
            face.append(self._position_of(point, face))
          self._faces.append(face)
    return self._faces

  def _position_of(self, point, face):
    """
    position of a specific point in the list of points that define a face
    :return: int
    """
    vertices = self.vertices
    for i in range(len(vertices)):
      # ensure not duplicated vertex
      power = 0
      vertex2 = vertices[i]
      for dimension in range(0, 3):
        power += math.pow(vertex2[dimension] - point[dimension], 2)
      distance = math.sqrt(power)
      if i not in face and distance == 0:
        return i
    return -1



"""
EnergySystem module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Peter Yefi peteryefi@gmail.com
"""

from hub.city_model_structure.city_object import CityObject
from hub.city_model_structure.energy_systems.air_source_hp import AirSourceHP
from hub.city_model_structure.energy_systems.water_to_water_hp import WaterToWaterHP


class EnergySystem(CityObject):
  """
  EnergySystem(CityObject) class
  """

  def __init__(self, name, surfaces):
    super().__init__(name, surfaces)
    self._air_source_hp = None
    self._water_to_water_hp = None
    self._type = 'energy_system'

  @property
  def air_source_hp(self) -> AirSourceHP:
    """
    Heat pump energy system
    :return:
    """
    return self._air_source_hp

  @air_source_hp.setter
  def air_source_hp(self, value):
    """
    Set heat pump for energy system
    :param value: AirSourceHP
    """
    if self._air_source_hp is None:
      self._air_source_hp = value

  @property
  def water_to_water_hp(self) -> WaterToWaterHP:
    """
    Water to water heat pump energy system
    :return:
    """
    return self._water_to_water_hp

  @water_to_water_hp.setter
  def water_to_water_hp(self, value):
    """
    Set water to water heat pump for energy system
    :param value: WaterToWaterHP
    """
    if self._water_to_water_hp is None:
      self._water_to_water_hp = value

  @property
  def type(self) -> str:
    """
    Type of city object
    :return: str
    """
    return self._type



"""
OsmSubway module parses osm files and import the metro location into the city model structure
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

import sys
import xmltodict
from pyproj import Transformer
from hub.city_model_structure.city import City
from hub.city_model_structure.subway_entrance import SubwayEntrance


class OsmSubway:
  """
  Open street map subway
  """
  def __init__(self, path):
    self._city = None
    self._subway_entrances = []
    with open(path) as osm:
      self._osm = xmltodict.parse(osm.read(), force_list='tag')
      for node in self._osm['osm']['node']:
        if 'tag' not in node:
          continue
        for tag in node['tag']:
          if '@v' not in tag:
            continue
          if tag['@v'] == 'subway_entrance':
            subway_entrance = SubwayEntrance(node['@id'], node['@lat'], node['@lon'])
            self._subway_entrances.append(subway_entrance)

  @property
  def city(self) -> City:
    """
    Get a city with subway entrances
    """
    transformer = Transformer.from_crs("EPSG:4326", "EPSG:3857")
    lower_corner = [sys.float_info.max, sys.float_info.max, 0]
    upper_corner = [sys.float_info.min, sys.float_info.min, 0]
    x = 0
    y = 1
    for subway_entrance in self._subway_entrances:
      coordinate = transformer.transform(subway_entrance.longitude, subway_entrance.latitude)
      if coordinate[x] >= upper_corner[x]:
        upper_corner[x] = coordinate[x]
      if coordinate[y] >= upper_corner[y]:
        upper_corner[y] = coordinate[y]
      if coordinate[x] < lower_corner[x]:
        lower_corner[x] = coordinate[x]
      if coordinate[y] < lower_corner[y]:
        lower_corner[y] = coordinate[y]

    city = City(lower_corner, upper_corner, 'unknown')
    for subway_entrance in self._subway_entrances:
      city.add_city_object(subway_entrance)
    return city
